/*  display.c -- Initialize and define display specifications for table

    See display.h for a description of the data structures involved.

    Written by Lee E. Brotzman, Hughes STX
    Written for Astronomical Data Center, NASA Goddard Space Flight Center

    This code is in the Public Domain, but no warranty, expressed or implied,
    is made by Hughes STX, NASA, or the ADC as to the accuracy and functioning
    of the software.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ftb.h"
#include "display.h"
#include "screen.h"
#include "table.h"
#include "utility.h"

/*
    Declare the display field structure; see DISPLAY.H for definition
*/
disp_t *display[MAXTABLES];


/*
    Local function definitions
*/
void Place(int tableno, int fldno, int position, int adjustment);
void PrintDisp(int tableno);


void Display(char *cmdline)
/*
    Parse a string which contains field display selections. A variable called
    currcol keeps track of the current display column.  The field selections
    are specified like so (fldid means a field identifier; either the field
    number or the field name):

    fldid           Display fldid at currcol;
                        currcol = currcol + width(fldid) + 1
    fldid1 fldid2   Display fields at currcol, separated by a blank
                        currcol = currcol + width(fldid1) + width(fldid2) + 1
    fldid1+fldid2   Display fields at currcol, no separation
                        currcol = currcol + width(fldid1) + width(fldid2)
    fldid/col       Display field at col;
                        currcol = col + width(fldid1) + 1
    fldid1, fldid2  Display fields fldid1 through fldid2, separated by blank
                        currcol = currcol + Sum((width(fldids)+1), n)

    Field spec:     fldid1+fldid2 fldid3/20 fldid4/7 fldid5, fldid6 fldid7
    width(fldid):   2      3      2         2        4       3      3

    Display: 111111111122222222223333333333444444444455555555556
    123456789012345678901234567890123456789012345678901234567890

    11222 44 5555 666  33 777
    | |   |  |    |    |  |
    +-|---|--|----|----|--|-- fldid1 at col 1;  currcol = currcol + 2     =  3
      +---|--|----|----|--|-- fldid2 at col 3;  currcol = currcol + 3 + 1 =  7
          |  |    |    +--|-- fldid3 at col 20; currcol = 20      + 2 + 1 = 23
          +--|----|-------|-- fldid4 at col 7;  currcol = 7       + 2 + 1 = 10
             +----|-------|-- fldid5 at col 10; currcol = currcol + 4 + 1 = 15
                  +-------|-- fldid6 at col 15; currcol = currcol + 3 + 1 = 19
                          +-- fldid7 at col 23;
     The next field specified will overwrite field 3 in column 15
*/
{   int     tableno;        /* Table number */
    char    object[81];     /* Temporary storage for field spec(s) */
    char    spec[81];       /* Individual parsed specification parameters */
    char    temp[81];       /* Temporary string holder */
    char    *special;       /* Pointer to special character */
    int     fldno;          /* Field number */
    int     fldno2;         /* Second field number in a range */
    int     colno;          /* Explicit column placement specification */
    int     i;              /* Loop control variable */

    SET.browse = FALSE;                 /* Clear browse mode */

    GetTableNo(cmdline, &tableno);
    scan(cmdline, ' ', object);
    /*
        Print current display settings if no other command entered.
    */
    if (object[0] == '\0') {
        PrintDisp(tableno);
        return;
    }
    /*
        Was a CLEAR command entered?  If so, erase the display information.
    */
    if (stricmp(object, "CLEAR") == 0) {
        FreeDisplay(tableno);
        if (strlen(cmdline) == 0)       /* Are we done? */
            return;                     /* If so, exit  */
        else
            scan(cmdline, ' ', object); /* Else, get next option */
    }

    /*
        Make sure a display entry exists for this table
    */
    if (display[tableno] == NULL)
        display[tableno] = NewDisplay();

    while(object[0] != '\0') {
        /*
            Look for a separator character; '+', '/', or ','
        */
        fldno = 0;
        special = strpbrk(object, "+/,");

        /*
            fldid separated by blanks
        */
        if (special == NULL) {
            FieldID(tableno, object, &fldno);
            if (fldno != 0)                   /* Place at current column and */
                Place(tableno, fldno, 0, 1);  /*     add a blank column      */
        }

        /*
            fldids separated by ',' -- place all fields in between
        */
        else if (*special == ',') {
            scan(object, ',', spec);
            FieldID(tableno, spec, &fldno);
            if (fldno != 0) {
                scan(object, ' ', spec);      /* Get second fldid in range   */
                if (*spec == '\0')            /* If blanks are in the way,   */
                    scan(cmdline, ' ', spec); /*    scan past them           */
                FieldID(tableno, spec, &fldno2);
                if (fldno2 != 0) {
                    if (fldno > fldno2) {     /* Adjust range if necessary */
                        int temp;
                        temp = fldno;
                        fldno = fldno2;
                        fldno2 = temp;
                    }
                    /* Place each field with with a blank separator */
                    for (i = fldno; i <= fldno2; i++)
                        Place(tableno, i, 0, 1);
                }
                else
                    strcpy(object, spec);
            }
            else strcpy(object, spec);
        }

        /*
            fldid separated by '+', place the two fields next to each other
        */
        else if (*special == '+') {
            scan(object, '+', spec);
            if (object[0] != '\0') {        /* Put any remainder back in   */
                strcpy(temp, cmdline);      /*    the selection list       */
                strcpy(cmdline, object); strcat(cmdline, " ");
                strcat(cmdline, temp);
            }
            FieldID(tableno, spec, &fldno);
            if (fldno != 0)                  /* Place at current column and */
                Place(tableno, fldno, 0, 0); /*   do not add a blank        */
            else
                strcpy(object, spec);
        }

        /*
            fldid separated by '/', place field at given column
        */
        else if (*special == '/') {
            scan(object, '/', spec);
            FieldID(tableno, spec, &fldno);
            colno = atoi(object);
            if (colno != 0 && fldno != 0)
                Place(tableno, fldno, colno, 1);
            else {
                ErrMsg("Invalid display column entered: %s", object);
                strcpy(object, spec);
            }
        }

        else
            ErrMsg("Invalid display specifier entered: %s", object);

        if (fldno == 0)
            ErrMsg("Invalid field identifier entered: %s", object);

        scan(cmdline, ' ', object);
    }
}


void Place(int tableno, int fldno, int position, int adjustment)
/*
    Place the field pointed to by 'fldno' into the display.  If 'position' = 0
    then put the field at the current display column, else put the field at the
    given position. 'adjustment' is an offset to add to the column number after
    the field has been placed.  The display column is incremented by the field
    width plus the adjustment.
*/
{   int pos, width;     /* Position and width of field to place in display */
    int MOVEIT;         /* Flag when a field won't fit at the display column */
    int i;

    table_t *T = table[tableno];        /* Shorthand */
    disp_t  *D = display[tableno];     /* Shorthand */

    /*
        First, make sure the field will fit at the column position
    */
    (position == 0) ? (pos = D->currcol) : (pos = position);
    width = T->field[fldno]->width;

    do {
        MOVEIT = FALSE;
        for (i = pos; i < (pos + width); i++)
            if (D->used[i] != '\0') {
                MOVEIT = TRUE;
                break;
            }
        if (MOVEIT)   /* Move the position to the next null field in 'used' */
            for (pos = i+1; (D->used[pos] != '\0') &&
                            (pos < (MAXDISP - width)); pos++);
    }   while(MOVEIT);

    /*
        Test for errors and/or warnings.  An error occurs if there is no room
        to place the field. A warning is issued if the expected display column
        had to be changed.
    */
    if (pos >= (MAXDISP - width)) {
        ErrMsg("Error: No room to display field %d, %s", fldno,
                T->field[fldno]->name);
        return;
    }
    if ((position == 0) && (pos != D->currcol)) {
        ErrMsg(
            "Warning: No room at current display column (%d), field %d, %s\n"
            "         will be displayed starting at column %d instead.",
            D->currcol, fldno, T->field[fldno]->name, pos);
    }
    if ((position != 0) && (pos != position)) {
        ErrMsg(
            "Warning: No room at chosen display column (%d), field %d, %s\n"
            "         will be displayed starting at column %d instead.",
            position, fldno, T->field[fldno]->name, pos);
    }

    D->count += 1;
    D->field[D->count].tabfield = fldno;
    D->field[D->count].dispcol  = pos;
    memset(&(D->used[pos]), fldno, width+adjustment);
    if (adjustment >= 0)
        D->currcol = pos + width + adjustment;
    if (D->currcol > D->width)
        D->width = D->currcol - adjustment - 1;
}


disp_t *NewDisplay(void)
/*
    Create and initialize a Display field entry pointed to by D.  READ THIS
    PROCEDURE CAREFULLY, for from this code certain assumptions about
    what the data structures look like are made.
*/
{   int i;
    disp_t *D;

    D = (disp_t *) malloc(sizeof(disp_t));
    D->count = 0;
    D->width = 0;
    for (i = 0; i < MAXFIELDS; i++) {
        D->field[i].tabfield = 0;
        D->field[i].dispcol  = 0;
    }
    D->currcol = 1;
    D->used    = (char *) malloc(MAXDISP+1);
    memset((void *) D->used, 0, MAXDISP+1);

    return D;
}


void FreeDisplay(int tableno)
/*
    Free and/or initialize all pieces of the display structure.  READ THIS
    FUNCTION CAREFULLY, for from this code certain assumptions about what the
    data structures look like are made.
*/
{   disp_t  *D = display[tableno];

    if (D != NULL) {
        if (D->used != NULL)
            free(D->used);
        free(D);
        display[tableno] = NULL;
    }
}


void CopyDisplay(int tableno)
/*
    Copy the field information from the table descriptors to the display
    descriptor.  The copy is one-to-one, resulting in the display appearing
    exactly as the input record.  The interaction of the table and display
    data structures are somewhat complex, and should be studied carefully.
*/
{   int     i;                      /* Loop control variable */
    table_t *T = table[tableno];    /* Shorthand */
    disp_t  *D = display[tableno]; /* Shorthand */

    D->count = T->fields;
    D->width = T->cols;
    D->currcol = T->field[T->fields]->column + T->field[T->fields]->width + 1;
    memset(D->used, '\0', T->cols);
    for (i = 1; i <= T->fields; i++) {
        D->field[i].tabfield = i;
        D->field[i].dispcol  = T->field[i]->column;
        memset(&(D->used[D->field[i].dispcol]), i, T->field[i]->width);
    }
}


void BuildDisplay(int tableno, long recno)
/*
     Build the display record, outrec (declared in module FTB.C), given the
     display information in the display descriptor.  Because of the complexity
     of the data structures involved, this procedure might take a while to
     figure out.  The source index is the 'column' parameter for the field
     pointed to in the display information array (for the given table).
*/
{   int  i;             /* Loop control variable */
    int  fldidx;        /* Temporary for holding display field index */
    int  source;        /* Index value of source string */
    int  dest;          /* Index value of destination string */
    int  count;         /* Number of bytes to move */

    table_t *T = table[tableno];    /* Shorthand */
    disp_t  *D = display[tableno];  /* Shorthand */

    if (SET.number)
        sprintf(outrec, "%06ld ", recno+1);

    for (i = 1; i <= D->count; i++) {
        fldidx = D->field[i].tabfield;
        source = T->field[fldidx]->column - 1;
        dest   = D->field[i].dispcol + SET.number - 1;
        count  = T->field[fldidx]->width;
        memcpy(&outrec[dest], &T->buffer[source], count);
    }
}


void PrintDisp(int tableno)
/*
    Print the field descriptors for the current display settings.
*/
{   int     i, j;           /* Loop control variables */
    char    outline[133];   /* Output line built from field information */

    disp_t  *D = display[tableno];  /* Shorthand */
    table_t *T = table[tableno];    /* Shorthand */
    field_t *F;                     /* Shorthand */

    if (D == NULL) {
        ErrMsg("WARNING: No display specifications for table %d (%s)\n",
               tableno, T->name);
        return;
    }

    Clear();
    Say("\nTable %-20s %s\n", T->name, T->namedesc);
    Say("\nFld Col  Fmt    Name            Units      Description\n\n");

    for (i = 1; i <= D->count; i++) {
        F = T->field[D->field[i].tabfield];
        sprintf(outline, "%3d%4d  %-7s", i, D->field[i].dispcol, F->format);

        if (F->name[0] != '\0')
            sprintf(&outline[strlen(outline)], "%-15s", trim(F->name));
        else
            sprintf(&outline[strlen(outline)], "%-15s", " ");

        if (F->units[0] != '\0')
            sprintf(&outline[strlen(outline)], " %-10s", trim(F->units));
        else
            sprintf(&outline[strlen(outline)], " %-10s", " ");

        if (F->namedesc[0] != '\0')
            sprintf(&outline[strlen(outline)], "%s", trim(F->namedesc));

        if (strlen(outline) < 80)
            Say("%s\n", outline);
        else {
            for (j = 79; outline[j] != ' '; --j)
                continue;
            outline[j] = '\0';
            ++j;
            Say("%s\n", outline);
            Say("%-44s%s\n", " ", &outline[j]);
        }
    }
}
