/*  extract.c -- Extract selected records into a disk file

    Written by Lee E. Brotzman, Hughes STX
    Written for Astronomical Data Center, NASA Goddard Space Flight Center

    This code is in the Public Domain, but no warranty, expressed or implied,
    is made by Hughes STX, NASA, or the ADC as to the accuracy and functioning
    of the software.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef MSDOS
#include <mem.h>
#endif
#ifdef unix
#include <memory.h>
#endif
#include "ftb.h"
#include "display.h"
#include "extract.h"
#include "fitsio.h"
#include "screen.h"
#include "table.h"
#include "utility.h"


/*
    Local storage
*/
char dashline[MAXDISP];
char nameline[MAXDISP];

/*
    Local function declarations
*/
void BuildBorder(int tableno);
void BuildDump(int tableno);
char *StringVal(char *value);
void WriteAny(FILE *outfile, ftb_t filetype, int length, char outrec[]);
void WriteComment(FILE *outfile, ftb_t filetype, char *keyword, comment_t *ptr);
void WriteHdr(FILE *outfile, ftb_t filetype, int tableno, long count);
void WriteNull(FILE *outfile, int tableno);


void Extract(char *cmdline)
/*
    Command: EXTRACT [TABLE table-id] [APPEND] [FITS|TEXT|DUMP]
                     [RANGE start-finish] [filename]

    Write display fields from selected records into the named file. The output
    can be in FITS mode, which will write a standard FITS ASCII table, TEXT
    mode, which will write a text version of the FITS header and a separate
    text file for the data, or DUMP mode, which dump the selected records with
    labelled column headings. The APPEND option in FITS mode will append an
    additional ASCII Table extension to the given FITS file, and in TEXT, or
    DUMP mode, will append data records to the given output text file.

    The RANGE search begins at record "start" and runs through record "finish"
    inclusive, the default is to start at record 1 and proceed to the end of
    the table.

    For filenames without extensions, EXTRACT will append standard extensions
    to the output filename(s). In FITS mode the extension is ".fit". In TEXT
    mode, the extensions are ".hdr" and ".dat" for the header and table file,
    respectively. In DUMP mode the extension is ".txt".
*/
{   FILE    *outfile;       /* Output file */
    char    filename[81];   /* Output file name */
    ftb_t   filetype;       /* Output file type (FITS or TEXT) */
    char    filemode[5];    /* Output file mode (write, update, or append) */
    char    *fileext;       /* Pointer to file extension */
    int     append;         /* Flag for file append */
    int     dump;           /* Flag for DUMP mode */
    char    option[81];     /* Temporary string for scanning command line */
    int     tableno;        /* Index of current table */
    long    start, finish;  /* Range for conducting the search */
    long    recno;          /* Current record number */
    long    count;          /* Count of records written */
    long    HeaderPos;      /* File pointer to beginning of table header */

    table_t *T;             /* Shorthand for table[tableno]    */
    disp_t  *D;             /* Shorthand for display[tableno] */

    SET.browse = FALSE;     /* Clear browse mode */

    outfile = NULL;
    filetype = FTB_FITS;
    strcpy(filemode, "w+");
    strcpy(filename, "");
    append  = FALSE;
    dump    = FALSE;
    recno   = 0L;
    start   = 1L;
    finish  = 0L;
    count   = 0L;

    GetTableNo(cmdline, &tableno);
    scan(cmdline, ' ', option);
    while (option[0] != '\0') {
        if (match(option, "FITS"))
            filetype = FTB_FITS;
        else if (match(option, "TEXT"))
            filetype = FTB_DAT;
        else if (match(option, "DUMP")) {
            filetype = FTB_TXT;
            dump = TRUE;
        }
        else if (match(option, "PRINT"))
            filetype = FTB_TXT;
        else if (match(option, "APPEND"))
            append = TRUE;
        else if (match(option, "RANGE")) {
            scan(cmdline, ' ', option);
            sscanf(option, "%ld-%ld", &start, &finish);
        }
        else
            strcpy(filename, option);
        scan(cmdline, ' ', option);
    }

    if (filename[0] == '\0')
        Prompt("Enter EXTRACT output filename: ", "", filename, 80);

    /*
        Put on appropriate filename extensions
    */
    if (strchr(filename, '.') == NULL) {
        if (filetype == FTB_FITS)
            strcat(filename, ".fit");
        else if (filetype == FTB_DAT)
            strcat(filename, ".dat");
        else if (filetype == FTB_TXT)
            strcat(filename, ".txt");
    }

    /*
        Set file mode, FITS is always written in update mode, others as streams
    */
    if (filetype == FTB_FITS)
        strcpy(filemode, append ? "r+" : "w+");
    else
        strcpy(filemode, append ? "a" : "w");

    if ((outfile = fopen(filename, filemode)) == NULL) {
        ErrMsg("ERROR opening file %s%s", filename,
                append ? " for append" : "");
        return;
    }

    /*
        Set up table and display pointers
    */
    T = table[tableno];
    D = display[tableno];
    if (D == NULL) {
        D = display[tableno] = NewDisplay();
        CopyDisplay(tableno);
    }

    /*
        Make sure the range is valid
    */
    if (start <= 0 || start > T->rows)
        start = 0;
    else
        start--;
    if (finish <= 0 || finish > T->rows)
        finish = T->rows;
    if (start > finish) {
        long temp = start;
        start = finish;
        finish = temp;
    }

    /*
        If we are processing FITS data and not appending an additional table
        extension write a null image header.
    */
    if (filetype == FTB_FITS) {
        if (!append)
            WriteNull(outfile, tableno);
        else
            fseek(outfile, 0L, SEEK_END);

        /*
            Write an ASCII table extension header for the current table. Field
            NAXIS2 can not be given yet, because we don't know how many records
            will be written, so ftell() is used to get the file position of the
            start of this header.  NAXIS2 is *always* the fifth record in the
            extension header, so the offset to the record is 'hard-coded'.
       */
        HeaderPos = ftell(outfile) + 320L;
        WriteHdr(outfile, filetype, tableno, count);
    }

    /*
        If we are processing in DUMP mode, build the appropriate dump file
        header records.
    */
    if (dump) {
        BuildBorder(tableno);
        WriteAny(outfile, filetype, 0, dashline);
        WriteAny(outfile, filetype, 0, nameline);
        WriteAny(outfile, filetype, 0, dashline);
    }

    /*
       Clean the output record
    */
    memset(outrec, ' ', D->width);
    outrec[D->width] = '\0';

    /*
        Start extracting table records
    */
    for (recno = start; recno < finish; recno++) {
        NextRec(tableno, start, finish, &recno);
        if (recno < finish) {
            if (dump)
                BuildDump(tableno);
            else
                BuildDisplay(tableno, recno);
            WriteAny(outfile, filetype, D->width, outrec);
            count++;
        }
    }

    /*
        If we are in FITS mode, we have to correct the value for NAXIS2, now
        that we know how many records were written.
    */
    if (filetype == FTB_FITS) {
        FlushFITS(outfile);
        fseek(outfile, HeaderPos, SEEK_SET);
        sprintf(outrec, "NAXIS2  = %20ld / Number of records in the table",
                count);
        fwrite(outrec, strlen(outrec), 1, outfile);
    }

    /*
        If we are in TEXT mode, we can now write a text file version of the
        FITS table header, since we know how many records have been written
        to the data file.  The output filename is that of the data file,
        with ".hdr" as the extension.
    */
    if (filetype == FTB_DAT) {
        filetype = FTB_HDR;
        if ((fileext = strrchr(filename, '.')) != NULL)
            strcpy(fileext, ".hdr");
        else
            strcat(filename, ".hdr");
        if ((outfile = freopen(filename, filemode, outfile)) == NULL) {
            ErrMsg("ERROR opening file %s", option);
            return;
        }
        WriteHdr(outfile, filetype, tableno, count);
    }

    /*
        If we are in DUMP mode, print out the final dashed line.
    */
    if (dump)
        WriteAny(outfile, filetype, 0, dashline);

    fclose(outfile);
}


void BuildBorder(int tableno)
/*
    Build a set of border records that will hold the field names
*/
{
    int  fldwidth;              /* Computed width of printed field */
    char *dashptr, *nameptr;    /* Pointers into the border record strings */
    char tmpbuf[81];
    int  i, j;                  /* Loop control variables */

    table_t *T = table[tableno];        /* Shorthand */
    disp_t  *D = display[tableno];      /* Shorthand */
    field_t *F;                         /* Shorthand */

    dashptr = dashline;
    nameptr = nameline;
    *dashptr++ = '+';
    *nameptr++ = '|';

    for (i = 1; i <= D->count; i++) {
        F = T->field[ D->field[i].tabfield ];
        fldwidth = max(F->width, strlen(F->name));
        sprintf(tmpbuf, "%-80s", F->name);
        for (j = 0; j < fldwidth; j++) {
            *dashptr++ = '-';
            *nameptr++ = tmpbuf[j];
        }
        *dashptr++ = '+';
        *nameptr++ = '|';
    }
    *dashptr = *nameptr = '\0';
}


void BuildDump(int tableno)
/*
    Build the output record, outrec, in print format.
*/
{
    int  fldwidth;  /* Computed width of printed field */
    char *print;    /* Pointer into the output print record */
    char *source;   /* Pointer into the source string */
    int  i, j;      /* Loop control variables */

    table_t *T = table[tableno];        /* Shorthand */
    disp_t  *D = display[tableno];      /* Shorthand */
    field_t *F;                         /* Shorthand */

    print = outrec;
    for (i = 1; i <= D->count; i++) {
        *print++ = '|';
        F = T->field[ D->field[i].tabfield ];
        fldwidth = max(F->width, strlen(F->name));
        source = &(T->buffer[ F->column - 1]);
        switch(F->format[0]) {
            case 'I' : case 'F' : case 'D' : case 'E' :
                if (fldwidth > F->width)
                    for (j = 0; j < fldwidth-(F->width); j++)
                        *print++ = ' ';
                for (j = 0; j < F->width; j++)
                    *print++ = *source++;
                break;
            case 'A' :
                for (j = 0; j < F->width; j++)
                    *print++ = *source++;
                for ( ; j < fldwidth; j++)
                    *print++ = ' ';
                break;
        }
    }
    *print++ = '|'; *print = '\0';
}


char *StringVal(char *value)
/*
    Print out a string in FITS delimited string form, i.e. enclosed in single
    quotes and at least 8 characters in length.  The function returns a pointer
    to a static string buffer.
*/
{
    static char valbuf[LONGSTR];    /* See TABLE.H */

    sprintf(valbuf, "\'%-8s\'", value);
    return valbuf;
}


void WriteAny(FILE *outfile, ftb_t filetype, int length, char outrec[])
/*
    Write a FITS header record to a FITS file or to a text file depending on
    the value of 'filetype'.
*/
{
    if (filetype == FTB_FITS)
        WriteFITS(outfile, length, outrec);
    else {
        fputs(outrec, outfile);
        if (outrec[strlen(outrec)] != '\n')
            fputs("\n", outfile);
    }
}


void WriteComment(FILE *outfile, ftb_t filetype, char *keyword, comment_t *ptr)
/*
    Write out a linked list of comments as COMMENT or HISTORY records.
*/
{
    comment_t *p;

    strcpy(outrec, keyword);
    p = ptr;
    do {
        p = p->next;
        outrec[strlen(keyword)] = '\0';
        strcat(outrec, p->data);
        WriteAny(outfile, filetype, 80, outrec);
    }   while (p != ptr);
}

void WriteHdr(FILE *outfile, ftb_t filetype, int tableno, long count)
/*
    Write FITS header records to the given file.  Function WriteAny() is used
    to control the type of file output based on 'filetype'.  All of the header
    information is extracted from the 'table' and 'display' structures (see
    TABLE.H and DISPLAY.H for details).  The only externally supplied data for
    the header is 'count', the number of records.  In FITS mode the header is
    written before the number of records has been determined and needs to be
    corrected afterwards.
*/
{
    int  i;             /* Loop control variable */
    char buff[81];      /* Temporary buffer for field comments */
    table_t *T = table[tableno];        /* Shorthand */
    disp_t  *D = display[tableno];      /* Shorthand */
    field_t *F;                         /* Shorthand */

    /*
        Start with the overall table description keywords
    */
    WriteAny(outfile, filetype, 80,
        "XTENSION= 'TABLE   '           / Table extension");
    WriteAny(outfile, filetype, 80,
        "BITPIX  =                    8 / Character data");
    WriteAny(outfile, filetype, 80,
        "NAXIS   =                    2 / Simple 2-D matrix");
    sprintf(outrec, "NAXIS1  = %20d / Number of characters per record",
        D->width);
    WriteAny(outfile, filetype, 80, outrec);
    sprintf(outrec, "NAXIS2  = %20ld / Number of records in the table",
        count);
    WriteAny(outfile, filetype, 80, outrec);
    WriteAny(outfile, filetype, 80,
        "PCOUNT  =                    0 / No \"random\" parameters");
    WriteAny(outfile, filetype, 80,
        "GCOUNT  =                    1 / Only one group");
    sprintf(outrec, "TFIELDS = %20d / Number of data fields per record",
        D->count);
    WriteAny(outfile, filetype, 80, outrec);
    if (strlen(T->name) > 0) {
        sprintf(outrec, "EXTNAME = %-20s / %s", StringVal(T->name),
                T->namedesc);
        WriteAny(outfile, filetype, 80, outrec);
    }
    if (strlen(T->author) > 0) {
        sprintf(outrec, "AUTHOR  = %-20s", StringVal(T->author));
        WriteAny(outfile, filetype, 80, outrec);
    }
    if (strlen(T->reference) > 0) {
        sprintf(outrec, "REFERENC= %-20s", StringVal(T->reference));
        WriteAny(outfile, filetype, 80, outrec);
    }
    if (strlen(T->date) > 0) {
        sprintf(outrec, "DATE    = %-20s / Date FITS table verified (dd/mm/yy)",
                StringVal(T->date));
        WriteAny(outfile, filetype, 80, outrec);
    }
    WriteAny(outfile, filetype, 80, " ");

    if (T->comment != NULL) {
        WriteComment(outfile, filetype, "COMMENT ", T->comment);
        WriteAny(outfile, filetype, 80, " ");
    }

    if (T->history != NULL) {
        WriteComment(outfile, filetype, "HISTORY ", T->history);
        WriteAny(outfile, filetype, 80, " ");
    }

    for (i = 1; i <= D->count; i++) {
        F = T->field[ D->field[i].tabfield ];
        if (strlen(F->name) > 0) {
            sprintf(outrec, "TTYPE%-3d= %-20s / %s", i, StringVal(F->name),
                    F->namedesc);
            WriteAny(outfile, filetype, 80, outrec);
        }
        sprintf(outrec, "TBCOL%-3d= %20d / Start column", i,
                D->field[i].dispcol);
        WriteAny(outfile, filetype, 80, outrec);
        sprintf(outrec, "TFORM%-3d= %-20s / Fortran format", i,
                StringVal(F->format));
        WriteAny(outfile, filetype, 80, outrec);
        if (strlen(F->units) > 0) {
            sprintf(outrec, "TUNIT%-3d= %-20s / %s", i, StringVal(F->units),
                    F->unitdesc);
            WriteAny(outfile, filetype, 80, outrec);
        }
        if (strlen(F->null) > 0) {
            sprintf(outrec, "TNULL%-3d= %-20s / Null value", i,
                    StringVal(F->null));
            WriteAny(outfile, filetype, 80, outrec);
        }
        if (F->scale != 1.0) {
            sprintf(outrec, "TSCAL%-3d= %20G / Scale factor", i, F->scale);
            WriteAny(outfile, filetype, 80, outrec);
        }
        if (F->zeropt != 0.0) {
            sprintf(outrec, "TZERO%-3d= %20G / Zero point", i, F->zeropt);
            WriteAny(outfile, filetype, 80, outrec);
        }
        if (F->comment != NULL) {
            sprintf(buff, "COMMENT %s: ", F->name);
            WriteComment(outfile, filetype, buff, F->comment);
        }
        WriteAny(outfile, filetype, 80, " ");
    }

    WriteAny(outfile, filetype, 80, "END");
    if (filetype == FTB_FITS)
        FlushFITS(outfile);
}

void WriteNull(FILE *outfile, int tableno)
/*
    Write the standard null image header that goes at the beginning of every
    FITS table.  Include the COMMENT cards from the original FITS file (if
    any).
*/
{
    comment_t *p;

    WriteFITS(outfile, 80,
        "SIMPLE  =                    T / Standard FITS format");
    WriteFITS(outfile, 80,
        "BITPIX  =                    8 / Character information");
    WriteFITS(outfile, 80,
        "NAXIS   =                    0 / No image data array present");
    WriteFITS(outfile, 80,
        "EXTEND  =                    T / There may be standard extensions");
    WriteFITS(outfile, 80,
        "BLOCKED =                    T / The file may be blocked");
    WriteFITS(outfile, 80,
        "ORIGIN  = 'ADC     '           / This file written at the ADC, NASA/GSFC");
    WriteFITS(outfile, 80,
        "DATE    = '23/09/91'           / Date file written (dd/mm/yy)");
    WriteFITS(outfile, 80, " ");

    if ((p = file[ table[tableno]->filenum ]->comment) != NULL) {
        WriteComment(outfile, FTB_FITS, "COMMENT ", p);
        WriteFITS(outfile, 80, " ");
    }
    WriteFITS(outfile, 80, "END");
    FlushFITS(outfile);
}
