/*  fitsio.c -- Read and write records to/from FITS table files

    FITS table file I/O.  Function ReadRec() is a general purpose reader of
    records in fixed-length files.  Portions of the file may have differing
    record lengths (like FITS files for instance). It is assumed that the file
    was opened in "binary" mode, so that trailing EOL characters are ignored
    during the read.

    NextRec() is the main function for reading table records based on the
    user's selection criteria (see SELECT.H and SELECT.C for details).

    WriteFITS() and FlushFITS() are used to write arbitrary character strings
    in FITS format.  After a sequence of calls to WriteFITS(), FlushFITS()
    must be called to ensure that the final FITS block is complete (2880 bytes
    long).

    Written by Lee E. Brotzman, Hughes STX
    Written for Astronomical Data Center, NASA Goddard Space Flight Center

    This code is in the Public Domain, but no warranty, expressed or implied,
    is made by Hughes STX, NASA, or the ADC as to the accuracy and functioning
    of the software.
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef MSDOS
#include <mem.h>
#endif
#ifdef unix
#include <memory.h>
#endif
#include "ftb.h"
#include "fitsio.h"
#include "screen.h"
#include "select.h"
#include "table.h"
#include "utility.h"

/*
    Local static data required to keep track of where the last I/O operation
    was performed and to track the record interval progress report.
*/
#define BLKLEN 16384
char    block[BLKLEN+1];    /* Input buffer for reading disk sectors */
long    lastblk   = -1;     /* Block number within the file */
FILE    *lastfile = NULL;   /* File pointer from last read */

char    fitsbuff[2*FITSLEN];/* FITS output buffer */
int     fitsptr = 0;        /* Pointer to current location in fitsbuff */


void ReadRec(FILE *fileptr, long base, int reclen, long recno, char *record)
/*
    Read data record number 'recno' into buffer 'record'.  'base' is the byte
    offset into for the first data record in file 'fileptr'; 'reclen' is the
    number of bytes in each data record.
*/
{   long byte_offset;   /* Byte offset of first record in the file */
    long blk_offset;    /* Disk block where record is stored */
    int  rec_offset;    /* Record offset within disk block */
    int  length;        /* Lengths of partial records that span blocks */

    /*
        Calculate the offsets into the file for this record
    */
    byte_offset = base + (recno * reclen);
    blk_offset  = byte_offset / BLKLEN;
    rec_offset  = byte_offset % BLKLEN;

    /*
        Get the disk block, if necessary
    */
    if ((lastfile != fileptr) || (lastblk != blk_offset)) {
        lastfile = fileptr;
        lastblk = blk_offset;
        fseek(fileptr, (blk_offset*BLKLEN), SEEK_SET);
        fread(block, BLKLEN, 1, fileptr);
    }

    /*
        If the record is completely contained in the disk block, then move
        it to the record buffer, otherwise move what is there, read the
        next disk block, and get the remainder of the table record.
    */
    if ((rec_offset + reclen - 1) <= BLKLEN)
        memcpy(record, &block[rec_offset], reclen);

    else {
        length = BLKLEN - rec_offset;
        memcpy(record, &block[rec_offset], length);
        ++lastblk;
        fread(block, BLKLEN, 1, fileptr);
        length = reclen - length;
        memcpy(&record[reclen-length], block, length);
    }
    record[reclen] = '\0';
}


void NextRec(int tableno, long minrecno, long maxrecno, long *recno)
/*
    Read in the next display record.  If records are being filtered, i.e. the
    SELECT command is in force, search for the next record that meets the
    search criteria.
*/
{   int     FOUND;          /* Boolean indicating a record is found */
    int     percent = 0;    /* Countup register for reporting percentage */
    int     search;         /* Search interval, in percentage */

    table_t *T = table[tableno];                /* Shorthand */
    FILE    *F = file[T->filenum]->fileptr;     /* Shorthand */

    percent = ((float)(*recno-minrecno+1) / (float)(maxrecno-minrecno)) * 100;
    StatusAt(SCRCOLS-15, "Search: %3d%%", percent < 100 ? percent : 100);

    FOUND = FALSE;
    while (*recno < maxrecno && !FOUND) {
        ReadRec(F, T->tabbase, T->cols, *recno, T->buffer);
        search = (((float) *recno / (float) (maxrecno-minrecno)) * 100.0) + 0.5;
        if (search != percent) {
            percent++;
            StatusAt(SCRCOLS-15, "Search: %3d%%", percent<100 ? percent : 100);
        }
        FOUND = Evaluate(tableno);
        if (!FOUND)
            *recno += 1;
    }
}


void WriteFITS(FILE *out, int len, char rec[])
{
    int diff;

    /* Write to current position in FITS buffer */
    strcpy(&fitsbuff[fitsptr], rec);

    /* Pad out record with ASCII blanks */
    for (diff = len - strlen(rec); diff > 0; diff--)
        fitsbuff[fitsptr+len-diff] = ' ';

    /* If block is full, write it and move excess to beginning of buffer */
    fitsptr += len;
    if (fitsptr > FITSLEN) {
        fwrite(fitsbuff, FITSLEN, 1, out);
        memcpy(fitsbuff, &fitsbuff[FITSLEN], fitsptr-FITSLEN);
        fitsptr -= FITSLEN;
    }
}

void FlushFITS(FILE *out)
{
    /* Pad out last buffer an flush it to disk */
    memset(&fitsbuff[fitsptr], ' ', FITSLEN - fitsptr);
    fwrite(fitsbuff, FITSLEN, 1, out);
    fitsptr = 0;
}
