/*  fitsscan.c -- Scan and decode FITS table headers

    The functions in this file scan the FITS file and decode all the table
    headers it finds.  First the file is checked to make sure it is a FITS
    file.  Succeeding in that, the algorithm searches for table extensions.
    When a table extension is found, the keywords are decoded into the
    associated fields of the table descriptor array.  The number of tables
    encountered and decoded is returned in 'tableno'.

    Written by Lee E. Brotzman, Hughes STX
    Written for Astronomical Data Center, NASA Goddard Space Flight Center

    This code is in the Public Domain, but no warranty, expressed or implied,
    is made by Hughes STX, NASA, or the ADC as to the accuracy and functioning
    of the software.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <setjmp.h>
#ifdef MSDOS
#include <io.h>
#endif
#ifdef unix
#include <unistd.h>
#endif
#include "ftb.h"
#include "fitsio.h"
#include "fitsscan.h"
#include "screen.h"
#include "table.h"
#include "utility.h"

/*
    Possible returned types of keywords.  NOTE: these are NOT the strings
    which identify the keywords, but user-defined data types.  To add
    keyword types, they should be declared as data types in this section AND
    have an entry in function 'initlookup'.
*/
enum keydefs {
    simple, bitpix, naxis, extend, origin, end, xtension, pcount, gcount,
    tfields, extname, author, referenc, date, history, comment,
    ttype, tbcol, tform, tunit, tnull, tscal, tzero, unknown
};

/*
    Structure for keyword lookup table entry
*/
struct keyentry {
    char data[9];
    enum keydefs keydef;
};

/*
    Define size of keyword and field lookup tables
*/
#define KEYTABSIZE 15
#define FLDTABSIZE 8

/*
    Keyword lookup table
*/
struct keyentry keyword[KEYTABSIZE] = {
    { "SIMPLE  ", simple   }, { "BITPIX  ", bitpix  }, { "EXTEND  ", extend   },
    { "ORIGIN  ", origin   }, { "END     ", end     }, { "XTENSION", xtension },
    { "PCOUNT  ", pcount   }, { "GCOUNT  ", gcount  }, { "TFIELDS ", tfields  },
    { "EXTNAME ", extname  }, { "AUTHOR  ", author  }, { "REFERENC", referenc },
    { "DATE    ", date     }, { "HISTORY ", history }, { "COMMENT ", comment  }
};

/*
    Field keyword lookup table
*/
struct keyentry fieldkey[FLDTABSIZE] = {
    { "NAXIS", naxis }, { "TTYPE", ttype }, { "TBCOL", tbcol },
    { "TFORM", tform }, { "TUNIT", tunit }, { "TNULL", tnull },
    { "TSCAL", tscal }, { "TZERO", tzero }
};


/*
    Local function calls
*/
void SkipImage(int filenum, long *nextheader);
void Decode(int filenum, int *tableno, long *nextheader);
void StoreField(int recno, field_t *F, enum keydefs keydef, char *hdrdata);
void StoreComment(table_t *T, int index, char *hdrdata);
void AbortScan(int recno, char *keyword, char *errmsg);
void ParseKey(char *hdrdata, enum keydefs *keydef, int *index);
void ParseBool(char *data, char *value, int *error);
void ParseInt(char *data, int *value, int *error);
void ParseLong(char *data, long *value, int *error);
void ParseFloat(char *data, double *value, int *error);
void ParseChar(char *data, char *value, int *error);
void ParseComm(char *data, char *value);
void CvtFmt(char *f77fmt, char *cfmt);
comment_t *AddComment(comment_t *comment, char *data);

/*
    Define jump buffer for error recovery
*/
jmp_buf scanjmp;


/*
    Define global error messages for parsing
*/
char ERRCHAR[]  = "must be a character string";
char ERRINT[]   = "must be an integer";
char ERRFLOAT[] = "must be a floating point number";
char ERRBOOL[]  = "must be TRUE";


void ScanFITS(int filenum)
{   long    nextheader = 0L;    /* File offset of next header to process */
    long    filesize;           /* Size of file in bytes */
    int     tableno;            /* Index of next free table structure */
    int     first = TRUE;       /* True the first time a table's found */

    /*
        First test for FITS:  length divisible by 2880
    */
    filesize = lseek(fileno(file[filenum]->fileptr), 0L, SEEK_END);
    /*
        Linux' fseek() (used in fitsio.c) seems to want the file pointer
        back where it was.
    */
#ifdef LINUX
    lseek(fileno(file[filenum]->fileptr), 0L, SEEK_SET);
#endif
    if ((filesize % FITSLENL) != 0) {
        ErrMsg("Input file is not FITS -- invalid record length");
        return;
    }

    if (setjmp(scanjmp) == 0) {
        /*
            Scan through main header and determine offset of next extension
        */
        SkipImage(filenum, &nextheader);

        /*
            Now scan through all extensions, filling in table data
        */
        while (nextheader < filesize) {
            Decode(filenum, &tableno, &nextheader);
            if (tableno >= 0) {             /* If a table was found... */
                file[filenum]->links++;     /* Bump the number of file links */
                TABLES++;                   /* Bump the number of tables */
                if (first) {                /* Mark the first table current */
                    first = FALSE;
                    SET.table = tableno;
                }
            }
        }
    }

    else {      /* There was a processing error, try to clean up last table */
        if ((tableno > 0) && (table[tableno] != NULL))
            FreeTable(tableno);
        for (tableno = 0; tableno < MAXTABLES; tableno++)
            if (table[tableno] != NULL) {
                SET.table = tableno;
                break;
            }
    }
}


void SkipImage(int filenum, long *nextheader)
/*
    Skip an image array in an attempt to scan through to the next table.
    'filenum' is the array index of the file data for the file currently being
    searched.  'hdrbase' is the base offset of the current header to skip.
    'nextheader' is the file offset of the next header to check.
*/
{
    enum keydefs keydef;    /* Used to parse keywords */
    int     index;          /* Field index returned from keywords like NAXIS1 */
    int     recno;          /* Header record number */
    char    hdrdata[81];    /* Parse buffer for header record */
    int     error;          /* Error flag from keyword parse routine */
    char    errmsg[81];     /* Error message text */
    struct {                /* Structure to hold image header information */
        char SIMPLE;
        int  BITPIX;
        long NAXIS[11];     /* NAXIS[0] holds value of NAXIS keyword */
        long PCOUNT;
        long GCOUNT;
        char EXTEND;
        long size;
    }   image = { 'F', 8, {0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L}, 0L, 1L, 'T', 0L};
    long    hdrsize;        /* Header size in bytes */

    file_t *F = file[filenum];       /* Shorthand file structure pointer */

    /*
        Decode the keywords of interest, namely: SIMPLE, BITPIX, NAXIS,
        NAXISn, PCOUNT, GCOUNT, and EXTEND.  Test for acceptable values
        while decoding.  Abort if bad values are found.

        Prime the pump by reading the first header record
    */
    recno = 0;
    ReadRec(F->fileptr, *nextheader, 80, recno++, hdrdata);

    ParseKey(hdrdata, &keydef, &index);

    while (keydef != end) {
        switch (keydef) {
            case simple:    /* Check for SIMPLE = T */
                ParseBool(hdrdata, &image.SIMPLE, &error);
                if (error || (image.SIMPLE == 'F'))
                    AbortScan(recno, "SIMPLE", ERRBOOL);
                break;

            case bitpix:    /* Check for BITPIX = n; where (n MOD 8) = 0 */
                ParseInt(hdrdata, &image.BITPIX, &error);
                if (error || (abs(image.BITPIX) % 8) != 0) {
                    strcpy(errmsg, ERRINT);
                    strcat(errmsg, " divisible by 8");
                    AbortScan(recno, "BITPIX", errmsg);
                }
                break;

            case naxis:
                if (index == 0) {   /* Parse NAXIS keyword */
                    ParseLong(hdrdata, &image.NAXIS[0], &error);
                    if (error || (image.NAXIS[0] > 10)) {
                        strcpy(errmsg, ERRINT);
                        strcat(errmsg, " <= 10");
                        AbortScan(recno, "NAXIS", errmsg);
                    }
                }
                else {              /* Parse NAXISn keyword(s) */
                    if (index > image.NAXIS[0]) {
                        strcpy(errmsg, ERRINT);
                        strcat(errmsg, " with n <= NAXIS");
                        AbortScan(recno, "NAXISn", errmsg);
                    }
                    ParseLong(hdrdata, &image.NAXIS[index], &error);
                    if (error)
                        AbortScan(recno, "NAXISn", ERRINT);
                }
                break;

            case pcount:    /* Parse PCOUNT */
                ParseLong(hdrdata, &image.PCOUNT, &error);
                if (error)
                    AbortScan(recno, "PCOUNT", ERRINT);
                break;

            case gcount:    /* Parse GCOUNT */
                ParseLong(hdrdata, &image.GCOUNT, &error);
                if (error)
                    AbortScan(recno, "GCOUNT", ERRINT);
                break;

            case extend:    /* Check for EXTEND = T */
                ParseBool(hdrdata, &image.EXTEND, &error);
                if (error || (image.EXTEND == 'F'))
                    AbortScan(recno, "EXTEND", ERRBOOL);
                break;

            case comment:   /* Parse COMMENT keyword */
                if (image.SIMPLE == 'T')
                    F->comment = AddComment(F->comment, hdrdata);
                break;

        }

        ReadRec(F->fileptr, *nextheader, 80, recno++, hdrdata);
        ParseKey(hdrdata, &keydef, &index);
    }

    /*
        Is this a Random Groups file?  If so, quit.
    */
    if (image.SIMPLE == 'T' && image.NAXIS[0] > 0 && image.NAXIS[1] == 0)
        AbortScan(3, "NAXIS1", "indicates Random Group format");

    /*
        Now calculate the size of the image which follows the main header
    */
    if (image.NAXIS[0] == 0)
        image.size = 0L;
    else {
        image.size = 1;
        for (index = 1; index <= image.NAXIS[0]; index++)
            image.size *= image.NAXIS[index];
        image.size += image.PCOUNT;
        image.size *= abs(image.BITPIX) * image.GCOUNT;
        image.size /= 8;
        /*
            Make the size divisible by FITSLENL
        */
        if ((image.size % FITSLENL) != 0)
            image.size = ((image.size / FITSLENL) + 1L) * FITSLENL;
    }

    /*
        Use the number of the record containing "END" to calculate the size of
        the header in bytes, remembering that the header block must be evenly
        divisible by 2880.  Add this to the image size and return the file
        offset of the next header.
    */
    hdrsize = ((long) --recno / 36L + 1L) * FITSLENL;
    *nextheader = *nextheader + hdrsize + image.size;
}


void Decode(int filenum, int *tableno, long *nextheader)
/*
    Decode FITS table header into the table data structure.

    If the table header pointed to by 'table[tableno].hdrbase' (see 'table'
    data structure in TABLE.H) is a valid FITS table header, i.e. EXTENSION =
    'TABLE', then decode the header into the table entry pointed to by
    'table[tableno]' and increment 'tableno' to index into the next table data
    array entry.  If the header is not for a table, then do not increment
    'tableno'.  In any event, calculate the sequential record number of the
    next FITS extension record in the file.
*/
{   enum keydefs keydef;    /* Used to parse keywords */
    int  index;             /* Field index */
    int  lastindex;         /* Backup copy of field index */
    int  recno;             /* Header record number */
    char hdrdata[81];       /* Parse buffer for header record */
    char temp[LONGSTR];     /* Temporary buffer */
    int  error;             /* Error flag from keyword parse routine */
    long tabsize;           /* Table size in bytes */
    long hdrsize;           /* Header size in bytes */

    table_t *T;                 /* Shorthand for table structure pointer */
    file_t  *F = file[filenum]; /* Shorthand for file structure pointer */

    /*
        Read the first record in the header and test if this is a table
    */
    recno = 0;
    ReadRec(F->fileptr, *nextheader, 80, recno++, hdrdata);
    ParseKey(hdrdata, &keydef, &index);
    ParseChar(hdrdata, temp, &error);

    if ((keydef != xtension) || (strcmp(trim(temp), "TABLE") != 0)) {
        SkipImage(filenum, nextheader);
        *tableno = -1;           /* Indicate that no table was found */
    }

    else {      /* A table is found, allocate space and begin decoding */
        for (*tableno = 1; *tableno < MAXTABLES; (*tableno)++)
            if (table[*tableno] == NULL) {
                if ((T = table[*tableno] = NewTable()) == NULL) {
                    ErrMsg("No more room in memory for table information");
                    *tableno = -1;
                    return;
                }
                break;
            }

        if (*tableno == MAXTABLES) {
            ErrMsg("Only %d tables can be open at one time.", MAXTABLES-1);
            *tableno = -1;
            return;
        }

        T->filenum = filenum;
        T->hdrbase = *nextheader;

        while (keydef != end) {
            ReadRec(F->fileptr, T->hdrbase, 80, recno++, hdrdata);
            ParseKey(hdrdata, &keydef, &index);

            switch (keydef) {
                case ttype: case tbcol: case tform: case tunit:
                case tnull: case tscal: case tzero:
                    if (T->field[index] == NULL)
                        if ((T->field[index] = NewField()) == NULL) {
                        ErrMsg("No more room in memory for field information.");
                            return;
                    }
                    StoreField(recno, T->field[index], keydef, hdrdata);
                    lastindex = index;
                    break;

                case comment:
                    StoreComment(T, lastindex, hdrdata);
                    break;

                case history:
                    T->history = AddComment(T->history, hdrdata);
                    break;

                case naxis:
                    if (index == 1) {
                        ParseInt(hdrdata, &T->cols, &error);
                        if (error)  AbortScan(recno, "NAXIS1", ERRINT);
                    }
                    else if (index == 2) {
                        ParseLong(hdrdata, &T->rows, &error);
                        if (error)  AbortScan(recno, "NAXIS2", ERRINT);
                    }
                    else if (index != 0)
                        AbortScan(recno, "NAXISn", "must be 2");
                    break;

                case tfields:
                    ParseInt(hdrdata, &T->fields, &error);
                    if (error)  AbortScan(recno, "TFIELDS", ERRINT);
                    break;

                case extname:
                    ParseChar(hdrdata, T->name, &error);
                    if (error)  AbortScan(recno, "EXTNAME", ERRCHAR);
                    trim(T->name);
                    ParseComm(hdrdata, T->namedesc);
                    break;

                case author:
                    ParseChar(hdrdata, T->author, &error);
                    if (error)  AbortScan(recno, "AUTHOR", ERRCHAR);
                    trim(T->author);
                    break;

                case referenc:
                    ParseChar(hdrdata, T->reference, &error);
                    if (error)  AbortScan(recno, "REFERENC", ERRCHAR);
                    trim(T->reference);
                    break;

                case date:
                    ParseChar(hdrdata, T->date, &error);
                    if (error)  AbortScan(recno, "DATE", ERRCHAR);
                    trim(T->date);
                    break;
            }
        }

        /*
            Allocate table record input buffer
        */
        T->buffer = (char *) malloc(MAXDISP);

        /*
            Calculate the size of the table, with padding for last FITS block
        */
        tabsize = T->cols * T->rows;
        if ((tabsize % FITSLENL) != 0)
            tabsize = ((tabsize / FITSLEN) + 1L) * FITSLENL;

        /*
            Use the number of the record containing "END" to calculate the size
            of the header in bytes, with padding to fill out the last block. Add
            this to the table size and return the offset of the next header.
        */
        hdrsize = ((long) --recno / 36L + 1L) * FITSLENL;
        *nextheader = T->hdrbase + hdrsize + tabsize;

        /*
            Store the file offset of the table data
        */
        T->tabbase = T->hdrbase + hdrsize;
    }
}


void StoreField(int recno, field_t *F, enum keydefs keydef, char *hdrdata)
/*
    Store the field information in hdrdata into the given field descriptor.
*/
{   int  scancode;
    int  error;

    switch (keydef) {
        case ttype:
            ParseChar(hdrdata, F->name, &error);
            if (error)  AbortScan(recno, "TTYPE", ERRCHAR);
            trim(F->name);
            ParseComm(hdrdata, F->namedesc);
            break;

        case tbcol:
            ParseInt(hdrdata, &F->column, &error);
            if (error)  AbortScan(recno, "TBCOL", ERRINT);
            break;

        case tform:
            ParseChar(hdrdata, F->format, &error);
            if (error)  AbortScan(recno, "TFORM", ERRCHAR);
            /*
                Determine width of field; make C format
            */
            trim(F->format);
            switch (F->format[0]) {
                case 'I': case 'A':
                    scancode = sscanf(&F->format[1], "%d.", &F->width);
                    if (scancode != 1)  AbortScan(recno, "TFORM",
                                        "format specifier is invalid");
                    break;
                case 'F': case 'E': case 'D':
                    scancode = sscanf(&F->format[1], "%d.%d", &F->width,
                                      &F->digits);
                    if (scancode != 2)  AbortScan(recno, "TFORM",
                                        "format specifier is invalid");
                    break;
            }
            CvtFmt(F->format, F->cfmt);
            break;

        case tunit:
            ParseChar(hdrdata, F->units, &error);
            if (error)  AbortScan(recno, "TUNIT", ERRCHAR);
            trim(F->units);
            ParseComm(hdrdata, F->unitdesc);
            break;

        case tnull:
            ParseChar(hdrdata, F->null, &error);
            if (error)  AbortScan(recno, "TNULL", ERRCHAR);
            break;

        case tscal:
            ParseFloat(hdrdata, &F->scale, &error);
            if (error)  AbortScan(recno, "TSCAL", ERRFLOAT);
            break;

        case tzero:
            ParseFloat(hdrdata, &F->zeropt, &error);
            if (error)  AbortScan(recno, "TZERO", ERRFLOAT);
            break;
    }
}


void StoreComment(table_t *T, int index, char *hdrdata)
/*
    For Astronomical Data Center FITS tables, comments about individual fields
    are written immediately following the field information and are encoded
    with the field's TTYPE followed by a colon, ':'.  For example:

    COMMENT RAH: This is a comment about field RAH

    This function looks at a comment record and checks to see if it belongs to
    the current field (pointed to by 'index'.  If it does, the comment text is
    added to the field information and the function returns TRUE, otherwise the
    comment text is add to the default table comment list.
*/
{
    char option[LONGSTR];
    char *backup;
    field_t *F = T->field[index];

    backup = strdup(hdrdata);
    scan(hdrdata, ':', option);
    if (stricmp(F->name, option) == 0)
        F->comment = AddComment(F->comment, hdrdata);
    else
        T->comment = AddComment(T->comment, backup);
    free(backup);
}


void AbortScan(int recno, char *keyword, char *errmsg)
/*
    Specialized abort routine for function 'decode'.  A customized error
    message is printed, along with its FITS block number and offset into the
    block.
*/
{   /*
        Print the block number and offset where the error occurred.
    */
    ErrMsg("Error detected in FITS file at record %d", recno);
    ErrMsg("Possible source of error: %s %s", keyword, errmsg);
    ErrMsg("The internal table structures may be corrupted, proceed with "
           "caution");
    longjmp(scanjmp,1);
}


void ParseKey(char *hdrdata, enum keydefs *keydef, int *index)
/*
    Parse the FITS keyword record in 'hdrdata' into a keyword type, determined
    by matching the keyword with a lookup table entry, an index if the
    keyword describes a field entry, and the data portion of the header
    record which follows the keyword (which is copied back into 'hdrdata').
*/
{   int i;      /* Loop index */
    int found;  /* Search flag */

    /*
        Field descriptors are more numerous, so check for them first
    */
    found = FALSE;
    for (i = 0; i < FLDTABSIZE; i++)
        if (strncmp(hdrdata, fieldkey[i].data, 5) == 0) {
            found   = TRUE;
            *keydef = fieldkey[i].keydef;
            *index  = atoi(&hdrdata[5]);
            break;
        }

    /*
        If a keyword wasn't found, search the other lookup table
    */
    if (!found)
        for (i = 0; i < KEYTABSIZE; i++)
            if (strncmp(hdrdata, keyword[i].data, 8) == 0) {
                found   = TRUE;
                *keydef = keyword[i].keydef;
                *index  = 0;
                break;
            }

    /*
        If the keyword STILL isn't found, then return null values
    */
    if (!found) {
        *keydef = unknown;
        index   = 0;
    }

    /*
        Modify the input string to point to the next data item. Skip the '='
        for those keywords that return a value.
    */
    if ((*keydef == comment) || (*keydef == history))
        strcpy(hdrdata, &hdrdata[8]);
    else
        strcpy(hdrdata, &hdrdata[10]);
}


void ParseBool(char *data, char *value, int *error)
/*
    Parse the data portion of a keyword string for the FITS boolean value
    'T' or 'F'.  The data string is modified to return any comment portion
    that might be present.  The character 'T' or 'F' is returned to
    indicate the value of the keyword.  If the keyword does not contain
    the correct type of data, 'error' is set to TRUE, otherwise FALSE.
*/
{   char temp[SHORTSTR];

    *error = FALSE;         /* Initialize error code */

    /*
        Scan for the value, which must equal 'T' or 'F'
    */
    scan(data, ' ', temp);
    if ((strcmp(temp, "T") != 0) && (strcmp(temp, "F") != 0))
        *error = TRUE;
    else
        *value = temp[0];
}


void ParseInt(char *data, int *value, int *error)
/*
    Parse the data portion of a keyword string for an integer keyword value.
    The data string is modified to return any comment portion that might
    be present.  If the keyword does not contain the correct type of data,
    'error' is set to TRUE, otherwise FALSE.
*/
{   char temp[SHORTSTR];
    int  scancode;

    *error = FALSE;         /* Initialize error code */

    /*
        Scan for the value, which must be an integer
    */
    scan(data, ' ', temp);
    scancode = sscanf(temp, "%d", value);
    if (scancode != 1)
        *error = TRUE;
}


void ParseLong(char *data, long *value, int *error)
/*
    Parse the data portion of a keyword string for an integer keyword value.
    The data string is modified to return any comment portion that might
    be present.  If the keyword does not contain the correct type of data,
    'error' is set to TRUE, otherwise FALSE.
*/
{   char temp[SHORTSTR];
    int  scancode;

    *error = FALSE;         /* Initialize error code */

    /*
        Scan for the value, which must be an integer
    */
    scan(data, ' ', temp);
    scancode = sscanf(temp, "%ld", value);
    if (scancode != 1)
        *error = TRUE;
}


void ParseFloat(char *data, double *value, int *error)
/*
    Parse the data portion of a keyword string for an floating point
    keyword value. The data string is modified to return any comment portion
    that might be present.  If the keyword does not contain the correct type
    of data, 'error' is set to TRUE, otherwise FALSE.
*/
{   char    temp[SHORTSTR];
    double  tvalue;
    int     scancode;

    *error = FALSE;         /* Initialize error code */

    /*
        Scan for the value, which must be a floating point number
        Note: The temporary floating point value is needed to avoid a runtime
                error message from scanf ('floating point formats not linked')
    */
    scan(data, ' ', temp);
    scancode = sscanf(temp, "%lf", &tvalue);
    if (scancode != 1)
        *error = TRUE;
    else
        *value = tvalue;
}


void ParseChar(char *data, char *value, int *error)
/*
    Parse the data portion of a keyword string for a character string keyword
    value delimited by single quotes.  The data string is modified to return
    any comment portion that might be present. If the keyword does not contain
    the correct type of data, 'error' is set to TRUE, otherwise FALSE.
*/
{   *error = FALSE;         /* Initialize error code */

    /*
        First, test to see if there is indeed a string to parse
    */
    if (data[0] != '\'')
        *error = TRUE;
    else    /* Scan for the value, which must be a character string */
        scan(data, '\'', value);
}


void ParseComm(char *data, char *value)
/*
    Parse the remainder of a keyword string for a comment following a '/'
    character.
*/
{   skip(data, ' ');             /* Skip any leading blanks */
    skip(data, '/');             /* Skip comment character */
    strcpy(value, trim(data));   /* Trim leading and trailing blanks */
}


void CvtFmt(char *f77fmt, char *cfmt)
/*
    Convert the given Fortran 77 format to an equivalent C format.
*/
{
    strcpy(cfmt, "%");
    if (f77fmt[0] == 'A')
        strcat(cfmt, "-");
    strcat(cfmt, &f77fmt[1]);
    trim(cfmt);
    switch (f77fmt[0]) {
        case 'A' : strcat(cfmt, "s"); break;
        case 'I' : strcat(cfmt, "i"); break;
        case 'F' : strcat(cfmt, "f"); break;
        case 'D' :
        case 'E' : strcat(cfmt, "e");
    }
}


comment_t *AddComment(comment_t *comment, char *data)
/*
    Add a comment or history line to the given list.  'comment' is a pointer
    to the last node in a circularly linked list.
*/
{   comment_t *P;

    /*
        First, check whether this is the first node for the list
    */
    if (comment == NULL) {
        comment = (comment_t *) malloc(sizeof(comment_t));
        strcpy(comment->data, trim(data));
        comment->next = comment;
        return comment;
    }

    /*
        For additional nodes, put data at end of the list
    */
    else {
        P = (comment_t *) malloc(sizeof(comment_t));
        strcpy(P->data, trim(data));    /* Store comment line */
        P->next = comment->next;        /* Last node points to first node */
        comment->next = P;              /* Next-to-last node points to last */
        return P;                       /* List pointer points to last */
    }
}
