/*  ftb.c -- Main program, command processor, miscellaneous commands

    This file contains the FTB main program, startup initialization routines,
    command processor, and a few minor commands.

    Written by Lee E. Brotzman, Hughes STX
    Written for Astronomical Data Center, NASA Goddard Space Flight Center

    This code is in the Public Domain, but no warranty, expressed or implied,
    is made by Hughes STX, NASA, or the ADC as to the accuracy and functioning
    of the software.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <setjmp.h>
#include <ctype.h>
#ifdef MSDOS
#include <dir.h>
#include <alloc.h>
#endif
#ifdef unix
#include <unistd.h>
#endif
#ifdef LINUX
#include <curses.h>
#endif
#include "ftb.h"
#include "display.h"
#include "extract.h"
#include "fitsio.h"
#include "list.h"
#include "open.h"
#include "screen.h"
#include "select.h"
#include "table.h"
#include "utility.h"


/*
    Global declarations
*/
set_t   SET;                            /* Global operational settings */
int     FILES  = 0;                     /* Current number of files */
int     TABLES = 0;                     /* Current number of active tables */
FILE    *helpfile;                      /* File containing help text */
char    helpfilename[13] = "ftb.hlp";   /* Default help file name  */
char    currdir[MAXCWD+1];              /* Current working directory */
long    recno;                          /* Current record number */
char    outrec[MAXDISP];                /* Output record for Browse/Extract */
jmp_buf CmdJmp;                         /* Jump buffer for signal interrupt */

/*
    Local storage for FTB.C
*/
char    homedir[MAXCWD+1];  /* Current directory when FTB was called */
int     homedisk;           /* Current disk when FTB was called */
jmp_buf CmdJmp;             /* setjmp/longjmp buffer for signal interrupt */

/*
    Local function prototypes for FTB.C
*/
void Banner(void);
void Browse(char *cmdline);
void Command(void);
void Handler(int sig);
void Help(char *cmdline);
void Initialize(void);
void Leave(void);
void Set(char *cmdline);
void Take(char *cmdline);
void Perform(char *cmdline);


void main(int argc, char *argv[])
{
    char option;            /* Command line option letter */
    int  i;                 /* Loop control variable */
    char cmdline[CMDLEN];   /* User command input */
    int  cmdend;            /* Counter holding the last position in cmdline */

    Initialize();           /* Perform one-time-only initializations */

    if (argc == 1) {        /* If there are no options given, display banner */
        InitScr();
        Banner();
    }

    else {
        /* Process command line options */
        for (i = 1; i < argc && (*argv[i] == '-' || *argv[i] == '/'); i++) {
                option = toupper(*++argv[i]);
                if (option == 'T') {    /* Option T: Execute a batch file */
                    if (*++argv[i] == '\0')
                        i++;
                    Take(argv[i]);
                }
         }
         if (i < argc) {
            InitScr();
            OpenFITS(argv[i]);
         }
    }

    /*
        Initialize I/O screens
    */
    if (!SET.curses)
        InitScr();

    /*
        Start infinite loop that executes interactive FTB commands
    */
    setjmp(CmdJmp);         /* A longjmp(CmdJmp) will always return here */
    for (;;) {
        Status("Path: %-25s  Files: %2d  Tables: %2d  Lines: %d/%d",
                currdir, FILES, TABLES, winrow+1, row);
        Scale();
        Prompt("FTB> ", "", cmdline, CMDLEN);
        cmdend = strlen(cmdline);
        while (cmdline[--cmdend] == SET.contchar) {
            Prompt("Continued> ", "", &cmdline[cmdend], CMDLEN-cmdend);
            cmdend = strlen(cmdline);
        }
        Perform(cmdline);
    }
}


void Perform(char *cmdline)
/*
    Perform whatever task was entered on the command line.
*/
{   char cmdbkp[CMDLEN];    /* Backup copy of the command line */
    char cmd[81];           /* Parsed command */
    int  tableno;

    if (strlen(trim(cmdline)) == 0)
        strcpy(cmdline, SET.defcmd);

/*    strcpy(SET.defcmd, cmdline); */
    strcpy(cmdbkp, cmdline);
    scan(cmdline, ' ', cmd);

    if (match(cmd, "BOTTOM"))
        Bottom(cmdline);

    else if (match(cmd, "BROWSE"))
        Browse(cmdline);

    else if (match(cmd, "CD")) {
        chdir(cmdline);
        getcwd(currdir, MAXCWD);
    }

    else if (match(cmd, "CLOSE")) {
        GetTableNo(cmdline, &tableno);
        FreeDisplay(tableno);
        FreeFile(table[tableno]->filenum);
        FreeTable(tableno);
    }

    else if (match(cmd, "DISPLAY"))
        Display(cmdline);

    else if (match(cmd, "DOWN"))
        Down(cmdline);

    else if (stricmp(cmd, "EXIT") == 0 || stricmp(cmd, "QUIT") == 0)
        Leave();

    else if (match(cmd, "EXTRACT"))
        Extract(cmdline);

    else if (match(cmd, "HELP"))
        Help(cmdline);

    else if (match(cmd, "LEFT"))
        Left(cmdline);

    else if (match(cmd, "LIST"))
        List(cmdline);

#ifdef MSDOS
    else if (match(cmd, "MEMORY"))
        Say("%lu bytes of RAM free\n", (unsigned long) coreleft());
#endif

    else if (match(cmd, "OPEN"))
        OpenFITS(cmdline);

    else if (match(cmd, "REMARK") || *cmd == '\0')
        { /* Do nothing for a remark or blank line */ }

    else if (match(cmd, "RIGHT"))
        Right(cmdline);

    else if (match(cmd, "SELECT"))
        Select(cmdline);

    else if (match(cmd, "SET"))
        Set(cmdline);

    else if (match(cmd, "SYSTEM"))
        System(cmdline);

    else if (match(cmd, "TAKE"))
        Take(cmdline);

    else if (match(cmd, "TOP"))
        Top(cmdline);

    else if (match(cmd, "UP"))
        Up(cmdline);

#ifdef MSDOS
    else if (strlen(cmdbkp) == 2 && isalpha(cmdbkp[0]) && cmdbkp[1] == ':') {
        setdisk(toupper(cmdbkp[0]) - 'A');
        getcwd(currdir, MAXCWD);
    }
#endif

    else if (strlen(cmdbkp) != 0)
        System(cmdbkp);
}


void Banner(void)
/*
    Display the opening screen for the FITS Browser.
*/
{
    Clear();
    Say("\n\n");
    Say("                   ADC FITS Table Browser, Version %s\n", VERNO);
    Say("\n                               Written by:\n");
    Say("\n                            Lee E. Brotzman\n");
    Say("                               Hughes STX\n");
    Say("\n                        Astronomical Data Center\n");
    Say("                     NASA Goddard Space Flight Center\n");
    Say("\n                 Work performed under Contract NAS5-30960\n");
    Say("\n              This program is Public Domain, but no warranty,\n");
    Say("            expressed or implied, is made by NASA or the ADC as\n");
    Say("              to the accuracy and functioning of the software.\n");
}


void Browse(char *cmdline)
/*
    Browse the table given on the command line
*/
{   int     tableno;        /* Index of current table */
    long    start  = 1;     /* Range for conducting the search */
    long    finish = 0;
    char    option[81];     /* Temporary string for scanning command line */
    long    toprec = -1;    /* Record number of first displayed record */
    long    newrec;         /* New record number to move to */
    int     count;          /* Running counter of records printed to screen */
    char    cmdbkp[CMDLEN]; /* Backup of command line buffer */
    char    cmd[21];        /* Parsed command */
    int     cmdend;

    SET.browse = TRUE;
    GetTableNo(cmdline, &tableno);              /* Get table identifier */

    scan(cmdline, ' ', option);                 /* Get search range */
    if (match(option, "RANGE")) {
        scan(cmdline, ' ', option);
        sscanf(option, "%ld-%ld", &start, &finish);
    }

    /*
        Make sure the range is valid
    */
    if (start <= 0 || start > table[tableno]->rows)
        start = 0;
    else
        start--;

    if (finish <= 0 || finish > table[tableno]->rows)
        finish = table[tableno]->rows;

    if (start > finish) {
        long temp = start;
        start = finish;
        finish = temp;
    }

    /*
        Set up display fields if that hasn't been done already
    */
    if (display[tableno] == NULL) {
        display[tableno] = NewDisplay();
        CopyDisplay(tableno);
    }

    /*
        Clean the output record; SET.number holds width of record number
    */
    memset(outrec, ' ', display[tableno]->width+SET.number);
    outrec[display[tableno]->width+SET.number] = '\0';

    /*
        Start displaying table records
    */
    Clear();
    newrec = start;
    while (SET.browse) {
        Status("Table: %2d. %-20s  Record: %ld/%ld",
                tableno, table[tableno]->name, newrec+1, finish);
        StatusAt(SCRCOLS-15, "Search: %3d%%", 0);
        Scale();

        /*
            Write table records
        */
        if (newrec != toprec) {
            recno = newrec;
            if (recno <  start)  recno = start;
            if (recno >= finish) recno = finish - 1;
            toprec = newrec = recno;

            At(0,0);
            for (count = 1; count < OUTLINES; count++, recno++) {
                /*
                   Seek the next display record.  NextRec() keeps looking
                   until a match is found, so 'recno' may be set to any
                   arbitrary value less then the total number of rows.
                   That's why it has to be checked against 'finish' again
                   after the call to NextRec().
                */
                if (recno < finish)
                    NextRec(tableno, start, finish, &recno);

                /* Display this data record */
                if (recno < finish) {
                    BuildDisplay(tableno, recno);
                    Say("%s\n", outrec);
                }
                else if (recno == finish)       /* Last record? */
                    Say("<End of table>\n");    /* If so, mark end of table... */
                else                            /* Past end of table... */
                    Say("\n");
                if (SET.update)
                    Update();
            }
        }

        Prompt("BROWSE> ", "", cmdline, CMDLEN);
        cmdend = strlen(cmdline);
        while (cmdline[--cmdend] == SET.contchar) {
            Prompt("Continued> ", "", &cmdline[cmdend], CMDLEN-cmdend);
            cmdend = strlen(cmdline);
        }

        if (strlen(trim(cmdline)) == 0)
            strcpy(cmdline, SET.defcmd);
        strcpy(cmdbkp, cmdline);
        scan(cmdline, ' ', cmd);

        if (match(cmd, "DOWN")) {
            int rows;
            if (match(cmdline, "PAGE"))
                rows = OUTLINES-1;
            else if (match(cmdline, "HALF"))
                rows = OUTLINES/2;
            else if ((rows = atoi(cmdline)) == 0)
                rows = OUTLINES/2;
            newrec = min(toprec+rows, finish-1);
        }
        else if (match(cmd, "UP")) {
            int rows;
            if (match(cmdline, "PAGE"))
                rows = OUTLINES-1;
            else if (match(cmdline, "HALF"))
                rows = OUTLINES/2;
            else if ((rows = atoi(cmdline)) == 0)
                rows = OUTLINES/2;
            newrec = max(toprec-rows, start);
        }
        else if (match(cmd, "TOP"))
            newrec = start;
        else if (match(cmd, "BOTTOM"))
            newrec = finish - OUTLINES + 2;
        else
           Perform(cmdbkp);
    }
}


void Handler(int sig)
/*
    Interrupt signal handler.  Includes fix for implicit noecho() call in some
    implementations of curses.
*/
{   char answer[11];        /* Response from user */
#ifndef echo
    int  echo();            /* Curses echo(), that is */
#endif

    Prompt("Do you wish to continue or quit? (C/Q)", "Continue", answer, 10);
    if (match(answer, "CONTINUE")) {
        if (SET.curses)
            echo();
        SET.browse = FALSE;
        signal(SIGINT, Handler);        /* Restore signal handler */
        longjmp(CmdJmp, 1);             /* Go back to Command() */
    }
    Leave();
}


void Help(char *topic)
/*
    Print help information on given topic.  The help is gotten from a text
    file called FTB.HLP.  Each topic is delimited by a record beginning
    with a period, '.', followed by the name of the topic.  All records up
    to the next topic heading are listed to the screen.
*/
{   char    inrec[81];      /* Records read from help file */
    int     found;          /* Boolean flag for whether a topic is found */

    SET.browse = FALSE;
    if (helpfile == NULL) { /* Make sure there is a help file to search */
        ErrMsg("Error: help file, %s, not found.", helpfilename);
        return;
    }

    if (strlen(topic) == 0)     /* Default help command is "HELP HELP" */
        strcpy(topic, "HELP");

    rewind(helpfile);

    found = FALSE;              /* Begin searching for topic */
    while(!feof(helpfile) && !found) {
        fgets(inrec, 80, helpfile);

        if (inrec[0] == '.') {
            /*
                Get rid of the newline character at the end of the record
            */
            inrec[strlen(inrec)-1] = '\0';
            if (match(topic, &inrec[1]))
                found = TRUE;
        }
    }


    if (!found)             /* If the given topic wasn't found do HELP HELP */
        Help("HELP");

    else {                  /*  otherwise list the help topic */
        Clear();
        fgets(inrec, 81, helpfile);
        while ((inrec[0] != '.') && !feof(helpfile)) {
            inrec[strlen(inrec)-1] = '\0';
            Say("%s\n", inrec);
            fgets(inrec, 81, helpfile);
        }
    }
}


void Initialize(void)
/*
    Initialize only the global data that has to be set once, not each time
    through the command loop.  READ THIS FUNCTION CAREFULLY for from this
    code certain assumptions are made about what the data look like.
*/
{   int  i;             /* Loop control variable */
    char filename[81];  /* temporary string to hold help file name */

    /*
        Initialize alphabetic names of logical operators
    */
    strcpy(OPS[L_AND],"and");     strcpy(OPS[L_OR],"or");
    strcpy(OPS[L_NOT],"!");
    strcpy(OPS[L_EQ],"=");        strcpy(OPS[L_NE],"!=");
    strcpy(OPS[L_LT],"<");        strcpy(OPS[L_GT],">");
    strcpy(OPS[L_GE],">=");       strcpy(OPS[L_LE],"<=");
    strcpy(OPS[L_SE],"SE");       strcpy(OPS[L_IN],"IN");
    strcpy(OPS[L_NONE],"NONE");

    /*
        Initialize operational settings
    */
    SET.browse  = FALSE;
    SET.contchar= '\\';
    SET.curses  = FALSE;
    strcpy(SET.defcmd, "DOWN HALF");
    SET.echo    = FALSE;
    SET.number  = 0;
    SET.table   = 0;
    SET.update  = FALSE;
    SET.wrap    = FALSE;

    /*
        Set the SIGINT handler.  Handler() asks the user whether to continue
    */
    signal(SIGINT, Handler);

    /*
        Make sure that all pointer variables start out as NULL
    */
    for(i = 0; i < MAXTABLES; i++) {
        table[i] = NULL;
        display[i] = NULL;
    }
    for(i = 0; i < MAXFILES; i++)
        file[i] = NULL;
    for(i = 0; i < MAXTEXT; i++)
        text[i] = NULL;

    /*
        Get current directory so it can be restored when the program exits.
    */
#ifdef MSDOS
    homedisk = getdisk();
#else
    homedisk = 0;
#endif
    getcwd(homedir, MAXCWD);
    strcpy(currdir, homedir);

    /*
        Open the help file, looking first in the current directory, then in
        the directory specified by the FTB environment variable and finally
        looking in the directories in the PATH.  See OPEN.C for details.
    */
    if ((helpfile = OpenFile(helpfilename, "r+t")) == NULL)
        ErrMsg("Warning:  help file %s not found.", filename);
}


void Leave(void)
/*
    Leave FTB, returning to the original disk and drive specifications.
*/
{
#ifdef MSDOS
    setdisk(homedisk);
#endif
    chdir(homedir);
    FreeScr();
    exit(0);
}


void Set(char *cmdline)
/*
    Set various global display and operational variables.
*/
{   char    option[CMDLEN];
    int     tableno;

    SET.browse = FALSE;                     /* Clear browse mode */
    scan(trim(cmdline), ' ', option);       /* Get option to SET */
    if (strlen(trim(option)) == 0) {        /* If none, list settings */
        Clear();
        Say("Current settings\n\n");
        Say("CONTINUE: %c\n", SET.contchar);
        Say("DEFAULT : %s\n", SET.defcmd);
        Say("ECHO    : %s\n", SET.echo   ? "ON" : "OFF");
        Say("NUMBER  : %s\n", SET.number ? "ON" : "OFF");
        Say("TABLE   : ");
        if (SET.table == 0)
            Say("No tables currently open\n");
        else if (table[SET.table]->name[0] == '\0')
            Say("%d. <No Name>\n", SET.table);
        else
            Say("%d. %s (%s)\n", SET.table, table[SET.table]->name,
                table[SET.table]->namedesc);
        Say("UPDATE  : %s\n", SET.update ? "ON" : "OFF");
        Say("WRAP    : %s\n", SET.wrap   ? "ON" : "OFF");
    }

    else if (match(option, "CONTINUE"))
        SET.contchar = cmdline[0];

    else if (match(option, "DEFAULT"))
        strcpy(SET.defcmd, strupr(cmdline));

    else if (match(option, "ECHO")) {
        if (stricmp(cmdline, "OFF") == 0)
            SET.echo = FALSE;
        else
            SET.echo = TRUE;
    }

    else if (match(option, "NUMBER")) {
        if (stricmp(cmdline, "ON") == 0)
            SET.number = 7;
        else
            SET.number = 0;
    }

    else if (match(option, "TABLE")) {
        if (cmdline[0] != '\0') {
            TableID(cmdline, &tableno);
            if (tableno)
                SET.table = tableno;
            else
                ErrMsg("ERROR: Invalid table identifier: %s", cmdline);
        }
    }
    else if (match(option, "UPDATE")) {
        if (stricmp(cmdline, "OFF") == 0)
            SET.update = FALSE;
        else
            SET.update = TRUE;
    }

    else if (match(option, "WRAP")) {
        if (stricmp(cmdline, "OFF") == 0)
            SET.wrap = FALSE;
        else
            SET.wrap = TRUE;
    }

    else
        ErrMsg("ERROR: Unrecognized SET option, %s", option);
}


void Take(char *cmdline)
/*
    Execute a batch file, which is simply a file containing a series of
    FTB commands.   Commands can be continued by putting a continuation
    character at the end of the line.  An error is generated if the command
    line is too long.
*/
{   FILE *takefile;     /* Batch file pointer */
    int  cmdend;        /* Pointer to end of the command line */

    SET.browse = FALSE;                 /* Clear browse mode */
    SET.batch  = TRUE;                  /* Set batch mode */
    if (strlen(trim(cmdline)) == 0)
        Prompt("Enter FTB batch file name: ", "", cmdline, CMDLEN);

    if ((takefile = fopen(trim(cmdline), "r+t")) == NULL) {
        ErrMsg("Error opening batch file %s", strupr(cmdline));
        return;
    }

    /*
        Get each command line from the batch file.  If a continuation character
        is found, then read in another line, appending it to the first.  No
        command line may be longer than CMDLEN characters.
    */
    cmdend = 0;
    while (fgets(&cmdline[cmdend], (CMDLEN-cmdend), takefile) != NULL) {
        cmdend = strlen(cmdline);
        cmdline[--cmdend] = '\0';
        if (cmdline[--cmdend] != SET.contchar) {
            if (SET.echo)
                Say("%s\n", cmdline);
            Perform(trim(cmdline));
            cmdend = 0;
        }
    }
    fclose(takefile);
    SET.batch = FALSE;                  /* Clear batch mode */
}
