/*  list.c -- List table information

    Print a human-readable listing of the header information for a given table.
    In all of the following, it is assumed that if a file pointer is passed
    as null then the output should go to the screen (it's done this way to
    allow something other than standard output to be used for writing to the
    screen, curses, for example).  A variation on Say(), called fSay(), is
    used to detect whether file redirection is necessary (see SCREEN.C).

    Written by Lee E. Brotzman, Hughes STX
    Written for Astronomical Data Center, NASA Goddard Space Flight Center

    This code is in the Public Domain, but no warranty, expressed or implied,
    is made by Hughes STX, NASA, or the ADC as to the accuracy and functioning
    of the software.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ftb.h"
#include "list.h"
#include "fitsio.h"
#include "screen.h"
#include "table.h"
#include "utility.h"


/*
    Local function prototypes
*/
void Tables(FILE *out);
void Header(FILE *out, int tableno);
void Info(FILE *out, int tableno);
void Fields(FILE *out, int tableno);
void Comments(FILE *out, comment_t *list, char *msg);


void List(char *cmdline)
/*
    List the active tables, or for a specific table, the table header, table
    information, or field information.  The default, if no options are given,
    is to list the active tables.  The output can be directed to stdout, stdprn,
    or to a file.   The default output device is the screen (stdout).
*/
{   int     tableno;        /* Current table number */
    char    option[81];     /* Parsed command line option */
    FILE    *out = NULL;    /* Output file for redirection */
    char    mode[2] = "w";  /* Output file mode ("w" or "a") */
    char    defname[81];    /* Name of default header information file */
    int     i = 0, j = 0;   /* Loop control variables */
    enum    { TABLES, HEADER, INFO, FIELDS }
            infotype[4] = { TABLES, -1, -1, -1};

    SET.browse = FALSE;             /* Clear browse mode */

    GetTableNo(cmdline, &tableno);
    if (table[tableno] == NULL)     /* Is there really a table here? */
        return;

    trim(cmdline);
    if (cmdline[0] == '\0')
        i = 1;
    else
        while (strlen(cmdline) > 0) {
            scan(cmdline, ' ', option);
            if (match(option, "HEADER"))
                infotype[i++] = HEADER;
            else if (match(option, "INFO"))
                infotype[i++] = INFO;
            else if (match(option, "FIELDS"))
                infotype[i++] = FIELDS;
            else if (match(option, "SCREEN"))
                out = NULL;
            else if (match(option, "APPEND"))
                strcpy(mode, "a");
            else if (match(option, "FILE")) {
                scan(cmdline, ' ', option);
                if (strlen(option) == 0) {
                    strcpy(defname, option);
                    Prompt("Enter file name to write to", defname, option, 80);
                }
                if ((out = fopen(option, mode)) == NULL)
                    ErrMsg("Error opening file %s for writing.", option);
            }
        }
    if (i == 0)
        i++;
    if (out == NULL)
        Clear();

    /*
        Now perform the action listed on the command line
    */
    for (j = 0; j < i; j++)
        switch (infotype[j]) {
            case TABLES:  Tables(out);           break;
            case HEADER:  Header(out, tableno);  break;
            case INFO:    Info  (out, tableno);  break;
            case FIELDS:  Fields(out, tableno);  break;
        }
    /*
        Clean up after the listings
    */
    if (out)
        fclose(out);                /* Flush a file by closing it */
}


void Tables(FILE *out)
/*
    Print a numbered list of the active tables
*/
{   int  i;                     /* Loop control variable */
    int  tablecnt, filecnt;     /* Number of active tables and files */

    for (tablecnt = i = 0; i < MAXTABLES; i++)
        if (table[i] != NULL)
            ++tablecnt;
    for (filecnt = i = 0; i < MAXFILES; i++)
        if (file[i]  != NULL)
            ++filecnt;

    fSay(out, "\nThere %s %d table%s in %d file%s",
            (tablecnt == 1) ? "is" : "are", tablecnt,
            (tablecnt == 1) ? ""   : "s",   filecnt,
            (filecnt  == 1) ? ""   : "s");
    fSay(out, "%s\n\n",
            (tablecnt != 0) ? " (Current table marked with an asterisk)" : "");

    for (tablecnt = i = 0; i < MAXTABLES; i++)
        if (table[i] != NULL) {
            fSay(out, "%c%2d. ", (i == SET.table) ? '*' : ' ', i);
            if (table[i]->name[0] == '\0')
                fSay(out, "<No Name>\n");
            else
                fSay(out, "%-20s %s\n", table[i]->name, table[i]->namedesc);
        }
}


void Header(FILE *out, int tableno)
/*
    Print the table header for the given table to the given output stream.
*/
{   char    headline[82];           /* Input buffer for header */
    long    recno = 0;              /* Input record number */
    table_t *T = table[tableno];              /* Shorthand */
    FILE    *F = file[T->filenum]->fileptr;   /* Shorthand */

    ReadRec(F, T->hdrbase, 80, recno, headline);
    trim(headline);
    do  {
        fSay(out, "%s\n", headline);
        ++recno;
        ReadRec(F, T->hdrbase, 80, recno, headline);
    }   while (!match(trim(headline), "END"));

    fSay(out, "%s\n", headline);
}


void Info(FILE *out, int tableno)
/*
    Print a human-readable listing of the given table descriptor to the given
    output device.
*/
{   table_t *T = table[tableno];      /* Shorthand */

    fSay(out, "\nName: %s, %s\n", T->name, T->namedesc);
    fSay(out, "File: %s\n", file[ T->filenum ]->filename);
    fSay(out, "Size: %ld rows, %d columns, %d fields\n\n",
              T->rows, T->cols, T->fields);

    if ((T->author[0] != '\0') || (T->reference[0] != '\0'))
        fSay(out, "Reference: %s %s\n", T->author, T->reference);

    if (T->history != NULL)
        Comments(out, T->history, "History:");

    if (T->comment != NULL)
        Comments(out, T->comment, "Comments:");
}


void Fields(FILE *out, int tableno)
/*
    Print the field descriptors for the given table descriptors to the given
    output stream.
*/
{   int     i, j;           /* Loop control variables */
    int     comments;       /* Boolean flag if there are any comments */
    char    note;           /* Note character indicating there are comments */
    char    outline[133];   /* Output line built from field information */

    table_t *T = table[tableno];    /* Shorthand */
    field_t *F;                     /* Shorthand */

    fSay(out, "\nTable %-20s %s\n", T->name, T->namedesc);
    fSay(out, "\nFld Col  Fmt    Name            Units      Description\n\n");

    comments = FALSE;
    for (i = 1; i <= T->fields; i++) {
        F = T->field[i];
        if (F->comment != NULL) {
            comments = TRUE;
            note = '*';
        }
        else
            note = ' ';

        sprintf(outline, "%3d%4d  %-7s", i, F->column, F->format);

        if (F->name[0] != '\0')
            sprintf(&outline[strlen(outline)], "%-15s", trim(F->name));
        else
            sprintf(&outline[strlen(outline)], "%-15s", " ");

        if (F->units[0] != '\0')
            sprintf(&outline[strlen(outline)], " %-10s", trim(F->units));
        else
            sprintf(&outline[strlen(outline)], " %-10s", " ");

        if (F->namedesc[0] != '\0')
            sprintf(&outline[strlen(outline)], "%c%s", note, trim(F->namedesc));

        if (strlen(outline) < 80)
            fSay(out, "%s\n", outline);
        else {
            for (j = 79; outline[j] != ' '; --j)
                continue;
            outline[j] = '\0';
            ++j;
            fSay(out, "%s\n", outline);
            fSay(out, "%-44s%s\n", " ", &outline[j]);
        }
    }
    if (comments)
        for (i = 1; i <= T->fields; i++)
            if (T->field[i]->comment != NULL) {
                strcpy(outline, T->field[i]->name);
                strcat(outline, ":");
                Comments(out, T->field[i]->comment, outline);
            }

}


void Comments(FILE *out, comment_t *list, char *msg)
/*
    Print the contents of a circularly-linked list which holds comment or
    history records.  'list' points to the last element in the circular list,
    'msg' is the string to be printed at the head of the list.
*/
{   comment_t *P;   /* Pointer to traverse comment/history lists */

    /*
        Write opening message
    */
    fSay(out, "\n%s\n", msg);

    /*
        Loop through the list from the first element to the last.  Stop
        when the last element printed or user says to stop.
    */
    P = list;
    do {
        P = P->next;
        fSay(out, "   %s\n", P->data);
    }   while (P != list);
}
