/*  open.c -- Open files

    OpenFITS opens the given FITS table file. The table file is analyzed to
    determine which type it is: standard FITS table, separate FITS header and
    data, or FITS binary table. Then the table file is run through the
    appropriate scan operation to get the table parameters.

    OpenFile is a generalized file opening function that looks for the file in
    (a) the current directory, (b) the directory pointed to by the "FTB"
    environment variable, and (c) the directories in the PATH.

    Written by Lee E. Brotzman, Hughes STX
    Written for Astronomical Data Center, NASA Goddard Space Flight Center

    This code is in the Public Domain, but no warranty, expressed or implied,
    is made by Hughes STX, NASA, or the ADC as to the accuracy and functioning
    of the software.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ftb.h"
#include "open.h"
#include "fitsscan.h"
#include "screen.h"
#include "table.h"
#include "utility.h"


void OpenFITS(char *cmdline)
/*
    Open a FITS table file.
*/
{   char  tempfilename[81];
    FILE *tempfile;
    int   filenum, i;
    comment_t *P;

    SET.browse = FALSE;             /* Clear browse mode */

    if (strlen(trim(cmdline)) == 0)
        Prompt("Enter table file name: ", "", cmdline, CMDLEN);

    scan(cmdline, ' ', tempfilename);

    /*
        Look to see if this file has been opened already, counting down from
        the top of the array.  The earliest NULL entry is tracked in case the
        file hasn't been opened.  If it is open, the file index is noted and
        the loop is broken.
    */
    for (filenum = -1, i = MAXFILES-1; i >= 0; i--) {
        if (file[i] == NULL)
            filenum = i;
        else if (stricmp(file[i]->filename, tempfilename) == 0) {
            filenum = i;
            file[i]->comment = FreeList(file[i]->comment);
            break;
        }
    }

    if (filenum < 0) {      /* Neither the file nor a slot for one was found */
        ErrMsg("No space available for more table files.");
        return;
    }
    else if (file[filenum] == NULL) {   /* An empty slot was found */
        if ((file[filenum] = NewFile()) == NULL) {
            ErrMsg("No more room in memory for another file.");
            return;
        }
        if ((tempfile = fopen(tempfilename, "rb")) == NULL) {
            FreeFile(filenum);
            ErrMsg("Error opening table file %s", tempfilename);
            return;
        }
        file[filenum]->filename = (char *) malloc(strlen(tempfilename)+1);
        strcpy(file[filenum]->filename, tempfilename);
        file[filenum]->fileptr  = tempfile;
        file[filenum]->filetype = FTB_FITS;
        file[filenum]->links = 0;
        file[filenum]->comment = NULL;
    }

    ScanFITS(filenum);

    if (file[filenum]->links == 0) {    /* No tables were found */
        ErrMsg("No tables were found in file %s", tempfilename);
        FreeFile(filenum);
    }
    /* If some tables were found, print any comments from main header */
    else if ((P = file[filenum]->comment) != NULL) {
        Clear();
        do {
            P = P->next;
            Say("   %s\n", P->data);
        }   while (P != file[filenum]->comment);
    }
}


FILE *OpenFile(char *filename, char *mode)
/*
    Open the given file using fopen() with the given mode.  This routine
    differs from fopen() in that is first tries to open the file in the current
    directory, and failing that tries the directory pointed to by the
    environment variable "FTB", and if *that* fails, it tries to find the file
    in the directories specified in the PATH environment variable.

    The temporary strings for the environment variables are stored in
    variable 'outrec' (see FTB.H for the definition of this character array).
*/
{
    FILE    *tmpfile;
    char    tmpfilename[256];
    char    *path;

    if ((tmpfile = fopen(filename, mode)) != NULL)
        return tmpfile;

    if ((path = getenv("FTB")) != NULL) {
        strcpy(tmpfilename, path);
        strcat(tmpfilename, PATHCHAR);
        strcat(tmpfilename, filename);
        if ((tmpfile = fopen(tmpfilename, mode)) != NULL)
            return tmpfile;
    }

    if ((path = getenv("PATH")) != NULL) {
        strcpy(outrec, path);
        path = strtok(outrec, PATHSEP);
        while (path != NULL) {
            strcpy(tmpfilename, path);
            strcat(tmpfilename, PATHCHAR);
            strcat(tmpfilename, filename);
            if ((tmpfile = fopen(tmpfilename, mode)) != NULL)
                return tmpfile;
            path = strtok(NULL, PATHSEP);
        }
    }

    return NULL;
}
