/*  readf.c -- Read Fortran-formatted floating point fields

    FITS specifies Fortran 77 formats for data stored in ASCII Tables.  This
    implies that data may be stored in formats readable by these formats.
    Fortran 77 reads are incompatible with C language reads (i.e. scanf())
    in that implicit decimal places are allowed in Fortran, but in C, implicit
    decimals result in erroneous read operations.  The following code was
    extracted from the "F2C" package written at AT&T Bell Laboratories and
    Bellcore (F2C copyright notice is included below).

    The only modifications to the subroutine were made to remove some
    generality, e.g. FITS specifies that reads are made as if the file was
    opened with BLANK=NULL, so references to the state of BLANK were changed
    to reflect this previous knowledge.

    Tests on several platforms show that floating point reads with readf()
    are roughly half as fast as the C library routine atof().  I find that
    perfectly acceptable, since readf() always returns the intended result.

    Changes from the original by Lee E. Brotzman, Hughes STX
    Written for Astronomical Data Center, NASA Goddard Space Flight Center
*/
/****************************************************************
Copyright 1990, 1991 by AT&T Bell Laboratories and Bellcore.

Permission to use, copy, modify, and distribute this software
and its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that the copyright notice and this
permission notice and warranty disclaimer appear in supporting
documentation, and that the names of AT&T Bell Laboratories or
Bellcore or any of their entities not be used in advertising or
publicity pertaining to distribution of the software without
specific, written prior permission.

AT&T and Bellcore disclaim all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall AT&T or Bellcore be liable for
any special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
****************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

/* FMAX = max number of nonzero digits passed to atof() */
/* EXPMAX = 10^EXPMAXDIGS - 1 = largest allowed exponent absolute value */
#define FMAX 40
#define EXPMAXDIGS 8
#define EXPMAX 99999999L

double readf(char *buf, int w, int d)
{
    char s[FMAX+EXPMAXDIGS+4];
    register int ch;
    register char *sp, *spe, *sp1;
    double x;
    int  se;
    long e, exp;
#ifndef atof
    double atof();
#endif

    sp1 = sp = s;
    spe = sp + FMAX;
    exp = -d;
    x = 0.;

    do {
        ch = *buf++;
        w--;
    }   while (ch == ' ' && w);
    switch(ch) {
        case '-': *sp++ = ch; sp1++; spe++;
        case '+':
            if (!w) goto zero;
            --w;
            ch = *buf++;
    }
    while(ch == ' ') {
        if (!w--) goto zero;
        ch = *buf++;
    }
    while(ch == '0') {
        if (!w--) goto zero;
        ch = *buf++;
    }
    while(isdigit(ch)) {
        if (sp < spe) *sp++ = ch;
        else ++exp;
digloop1e:
        if (!w--) goto done;
        ch = *buf++;
    }
    if (ch == ' ')
        goto digloop1e;
    if (ch == '.') {
        exp += d;
        if (!w--) goto done;
        ch = *buf++;
        if (sp == sp1) { /* no digits yet */
            while(ch == '0') {
                --exp;
skip0:
                if (!w--) goto done;
                ch = *buf++;
            }
            if (ch == ' ') goto skip0;
        }
        while(isdigit(ch)) {
            if (sp < spe) {
                *sp++ = ch;
                --exp;
            }
digloop2e:
            if (!w--) goto done;
            ch = *buf++;
        }
        if (ch == ' ') goto digloop2e;
    }
    switch(ch) {
        default:  break;
        case '-': se = 1; goto signonly;
        case '+': se = 0; goto signonly;
        case 'e':
        case 'E':
        case 'd':
        case 'D':
            if (!w--) goto bad;
            ch = *buf++;
            while(ch == ' ') {
                if (!w--) goto bad;
                ch = *buf++;
            }
            se = 0;
            switch(ch) {
                case '-': se = 1;
                case '+':
signonly:
                if (!w--) goto bad;
                ch = *buf++;
            }
            while(ch == ' ') {
                if (!w--) goto bad;
                ch = *buf++;
            }
            if (!isdigit(ch)) goto bad;

            e = ch - '0';
            for(;;) {
                if (!w--) {
                    ch = '\n';
                    break;
                }
                ch = *buf++;
                if (!isdigit(ch)) {
                    if (ch == ' ')
                        continue;
                    else
                        break;
                }
                e = 10*e + ch - '0';
                if (e > EXPMAX && sp > sp1) goto bad;
            }
            if (se)
                exp -= e;
            else
                exp += e;
    }
    switch(ch) {
        case '\n':
        case ',':
            break;
        default:
bad:
            return (0.0);
    }
done:
    if (sp > sp1) {
        while(*--sp == '0')
            ++exp;
        if (exp)
            sprintf(sp+1, "e%ld", exp);
        else
            sp[1] = 0;
        x = atof(s);
    }
zero:
    return(x);
}


