/*  screen.c -- Basic screen I/O

    This file contains the functions for moving the cursor, printing data in
    the output window, printing informational messages in the command window
    and getting user input. This is the only module that is allowed direct
    access to the "curses" screen manipulation functions.

    Written by Lee E. Brotzman, Hughes STX
    Written for Astronomical Data Center, NASA Goddard Space Flight Center

    This code is in the Public Domain, but no warranty, expressed or implied,
    is made by Hughes STX, NASA, or the ADC as to the accuracy and functioning
    of the software.
*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#ifndef ultrix
#include <curses.h>
#else
#include <cursesX.h>
#endif
#ifdef unix
#include <unistd.h>
#endif
#include "ftb.h"
#include "screen.h"
#include "utility.h"

/*
    Global declarations: specifications of the size of the display windows
*/
int     SCRLINES;        /* Number of screen lines */
int     SCRCOLS;         /* Number of screen columns */
int     OUTLINES;        /* Number of lines in output window */
int     row;             /* Pointer to next record to fill in the text array */
int     col;             /* Pointer to current column in text array */
int     winrow;          /* Row to display starting in upper left corner */
int     wincol;          /* Column to display starting in upper left corner */

/*
    Local data and function declarations
*/
WINDOW *outwin;         /* Command output window */
WINDOW *cmdwin;         /* Command input window  */
char   scratch[1000];   /* Scratch buffer for building output string */
char   *text[MAXTEXT];  /* Array of pointers to screen text */
int    touched;         /* Flag indicating whether the screen needs updating */


void At(int y, int x)
{
    row = y;
    col = x;
}


void Clear(void)
{
    row = col = winrow = wincol = 0;
    touched = TRUE;
    if (SET.curses) {
        werase(outwin);
        wmove(outwin,0,0);
        wrefresh(outwin);
    }
}


void ErrMsg(char *fmt, ...)
/*
    Print an error message in the command window
*/
{   va_list ap;
    va_start(ap,fmt);
    vsprintf(scratch, fmt, ap);
    if (!SET.curses)
        fprintf(stderr, "\a%s\n", scratch);
    else {
        wmove(cmdwin, 2, 0);
#ifdef SYSV
#ifndef LINUX
        wattron(cmdwin, A_STANDOUT);
#endif
#endif
        wclrtoeol(cmdwin);
        waddstr(cmdwin, scratch);
        wrefresh(cmdwin);
        wgetch(cmdwin);
#ifdef SYSV
#ifndef LINUX
        wattrset(cmdwin, A_NORMAL);
#endif
#endif
    }
    va_end(ap);
}


void FreeScr(void)
{
    if (SET.curses)
        endwin();
}


void Get(char *string)
/*
    Get a string from the command window
*/
{
    if (!SET.curses)
        gets(string);
    else {
        wgetstr(cmdwin, string);
        wrefresh(cmdwin);
    }
}


void InitScr(void)
{
    SET.curses = TRUE;
    initscr();
    SCRLINES = LINES;
    SCRCOLS  = COLS;
    OUTLINES = SCRLINES-3;
    row = col = winrow = wincol = 0;
    touched = TRUE;
    outwin = newwin(OUTLINES,COLS,LINES-OUTLINES,0);
    cmdwin = newwin(LINES-OUTLINES,COLS,0,0);
#ifdef SYSV
#ifndef LINUX
    wattrset(outwin, A_NORMAL);
    wattrset(cmdwin, A_NORMAL);
#endif
#endif
    scrollok(outwin, TRUE);
}


void Prompt(char *message, char *defresp, char *response, int resplen)
/*
    Print a message to the user, indicate the default response in square
    brackets ("[]"), and wait for a reponse.  If the user responds with a null
    string, then return the default, otherwise return the user's input string.

    The response is always returned with leading and trailing blanks removed.
    The case of the response is left intact.    No more than 'resplen' characters
    are read from stdin.
*/
{
    if (!SET.curses)
        InitScr();

    /*
        Update the screen
    */
    Update();

    /*
        Issue prompt message; get response from user
    */
    wmove(cmdwin, 1, 0);
    if (*defresp == '\0')
        wprintw(cmdwin, "%s ", message);
    else
        wprintw(cmdwin, "%s [%s]: ", message, defresp);
    wclrtoeol(cmdwin);
    wrefresh(cmdwin);

    Get(scratch);
    if (strlen(scratch) > resplen) {
        strncpy(response, scratch, resplen);
        response[resplen] = '\0';
    }
    else
        strcpy(response, scratch);

    /*
        If response was null string, simply return default
    */
    if (*response == '\0')
        strcpy(response, defresp);

    trim(response);
}


void fSay(FILE *out, char *fmt, ...)
/*
    Either use Say() to put the message into the screen area if the FILE
    pointer is NULL or else write it to the specified file.
*/
{   va_list ap;
    char scratch[250];

    va_start(ap,fmt);
    vsprintf(scratch, fmt, ap);
    if (out)
        fputs(scratch, out);
    else
        Say("%s", scratch);
    va_end(ap);
}


void Say(char *fmt, ...)
/*
    Put a message in the displayed text array
*/
{   int i;
    int last;
    char *first;
    va_list ap;
    va_start(ap,fmt);

    /*
        Copy the current record into a scratch buffer for modification, and
        pad with blanks out to the current column.
    */
    strncpy(scratch, text[row], col);
    for (i = strlen(scratch); i < col; i++)
        scratch[i] = ' ';
    scratch[col] = '\0';

    /*
        Copy the modified record(s) back to the heap.  The formatted string
        is checked for newline characters.  Each newline indicates a new
        record for the list.  Special care must be taken for records that
        do not end in a newline.
    */
    vsprintf(&scratch[col], fmt, ap);
    if (!SET.curses) {
        if (!SET.batch || (SET.batch && SET.echo))
            printf("%s", scratch);
    }
    else {
        first = scratch;
        last  = strcspn(first, "\n");
        /*
            In WRAP mode, no line is allowed to be longer than the screen
            width, and breaks are made at white space boundaries.
        */
        if (SET.wrap && last > SCRCOLS) {
            for (last = SCRCOLS; last && !isspace(first[last]); last--)
                continue;
            for ( ; last && isspace(first[last]); last--)
                continue;
            if (last)                    /* Was there a word break anywhere? */
                first[++last] = '\n';    /* If so, insert a newline */
        }

        /*
            Write the record to the internal text file
        */
        while (*first != '\0') {
            /* The array is allocated his way to keep GNU C happy on Suns
               and Linux boxes.  Ordinarily, I'd just use realloc() itself. */
            if (text[row] == NULL)
                text[row] = (char *) malloc(last+2);
            else
                text[row] = (char *) realloc(text[row], last+2);
            if (text[row] == NULL) {
                ErrMsg("Error in Say(): memory allocation");
                return;
            }
            strncpy(text[row], first, last+1);
            text[row][last+1] = '\0';
            if (first[last] == '\n') {
                first = &first[last+1];
                last = strcspn(first, "\n");
                row++;
                col = 0;
            }
            else {
                col = last;
                break;
            }
        }
        touched = TRUE;
    }
    va_end(ap);
}


void Scale(void)
/*
    Print 80-byte scale across screen.
*/
{   char data[500];
    int  offset;
    int  i, j;

    offset = 0;
    if (SET.browse && SET.number) {
        strcpy(data, "Record ");
        offset = SET.number;
    }

    for (i = offset, j = wincol+1; i < SCRCOLS; i++, j++) {
        data[i] = '-';
        if ((j/5*5) == j)       /* Divisible by 5, put a '+' */
            data[i] = '+';
        if ((j/10*10) == j) {   /* Divisible by 10, write the column */
            if (j < 100)
                sprintf(&data[i], "%1d", j/10);
            else if (j < 1000)
                sprintf(&data[i-1], "%2d", j/10);
            else
                sprintf(&data[i-2], "%3d", j/10);
        }
    }
    data[SCRCOLS+1] = '\0';
    if (!SET.curses)
        printf("%s", data);
    else {
        wmove(cmdwin, 2, 0);
        waddstr(cmdwin, data);
        wrefresh(cmdwin);
    }
}


void Status(char *fmt, ...)
/*
    Print the status line on the top of the screen.  Status items include the
    current working directory and the number of open files and tables.
*/
{   va_list ap;
    va_start(ap,fmt);
    if (SET.curses) {
        vsprintf(scratch, fmt, ap);
        wmove(cmdwin, 0, 0);
        wclrtoeol(cmdwin);
        waddstr(cmdwin, scratch);
        wrefresh(cmdwin);
    }
    va_end(ap);
}


void StatusAt(int x, char *fmt, ...)
{   va_list ap;
    va_start(ap,fmt);
    if (SET.curses) {
        vsprintf(scratch, fmt, ap);
        wmove(cmdwin, 0, x);
        waddstr(cmdwin, scratch);
        wrefresh(cmdwin);
    }
    va_end(ap);
}


void System(char *cmdline)
/*
    Pass the command line on to the operating system.   If the cmdline is blank
    or null, shell out to the operating system. The user must then use the
    "exit" OS command to return to FTB.
*/
{   if (SET.curses) {
        clear(); refresh();
        wmove(stdscr, OUTLINES, 0);
#ifdef SYSV
#ifndef LINUX
        reset_shell_mode();
#endif
#endif
    }
    trim(cmdline);
    if (cmdline[0] == '\0')
        puts("\nEnter the command EXIT to return to FTB.");
    system(cmdline);
    puts("\nPress ENTER to return to FTB: ");
    fflush(stdout);
    gets(cmdline);
    if (SET.curses) {
        clear(); refresh();
#ifdef SYSV
#ifndef LINUX
        reset_prog_mode();
#endif
#endif
        touchwin(cmdwin); wrefresh(cmdwin);
        touchwin(outwin); wrefresh(outwin);
    }
}


void Update(void)
/*
    Update the display window if any of the on-screen elements have changed.
    A little care must be taken to make sure that rows of data are still
    displayed correctly after the display has been scrolled left or right.
*/
{
    int i, line;
    int offset;

    if (!touched) return;

    if (SET.curses)
        wmove(outwin,0,0);
    for (i = 0, line = winrow; i < OUTLINES-1; i++, line++) {
        if (line < row && text[line] != NULL) {
            strcpy(scratch, text[line]);
            if (wincol) {
                offset = 0;
                if (SET.browse && SET.number)
                    offset = SET.number;
                if (wincol < strlen(scratch))
                    strncpy(&scratch[offset], &scratch[wincol+offset],
                            SCRCOLS-offset);
                else
                    strcpy(scratch, "\n");
            }
            scratch[SCRCOLS] = '\0';
        }
        else
            strcpy(scratch, "\n");
        if (!SET.curses)
            printf("%s", scratch);
        else {
            wclrtoeol(outwin);
            waddstr(outwin, scratch);
            wrefresh(outwin);
        }
    }
    touched = FALSE;
}


void Where(int *y, int *x)
{
    *y = row;
    *x = col;
}


void Top(char *cmdline)
/*
    Move the row pointer to the top of the screen text.
*/
{
    winrow = 0;
    touched = TRUE;
}


void Bottom(char *cmdline)
/*
    Move the row pointer to the bottom of the screen text.
*/
{
    winrow = row - OUTLINES + 2;
    touched = TRUE;
}


void Down(char *cmdline)
/*
    Move the row pointer by the number of lines given on the command line.
*/
{   int rows;

    if (winrow+OUTLINES > row)
        return;

    if (match(cmdline, "PAGE"))
        rows = OUTLINES-1;
    else if (match(cmdline, "HALF"))
        rows = OUTLINES/2;
    else if ((rows = atoi(cmdline)) == 0)
        rows = OUTLINES/2;
    winrow += rows;
    if (winrow < 0 || winrow > row)
        winrow = 0;
    touched = TRUE;
}


void Up(char *cmdline)
/*
    Move the row pointer by the number of lines given on the command line.
*/
{   int rows;

    if (match(cmdline, "PAGE"))
        rows = OUTLINES - 1;
    else if (match(cmdline, "HALF"))
        rows = OUTLINES/2;
    else if ((rows = atoi(cmdline)) == 0)
        rows = OUTLINES/2;
    winrow -= rows;
    if (winrow < 0)
        winrow = 0;
    if (winrow > row)
        winrow = row;
    touched = TRUE;
}


void Left(char *cmdline)
/*
    Move the column pointer by the number of columns given on the command line.
*/
{   int cols;

    if (match(cmdline, "PAGE"))
        cols = SCRCOLS - 1;
    else if (match(cmdline, "HALF"))
        cols = SCRCOLS/2;
    else if ((cols = atoi(cmdline)) == 0)
        cols = SCRCOLS/2;
    wincol -= cols;
    if (wincol < 0)
        wincol = 0;
    touched = TRUE;
}


void Right(char *cmdline)
/*
    Move the column pointer by the number of columns given on the command line.
*/
{   int cols;

    if (match(cmdline, "PAGE"))
        cols = SCRCOLS - 1;
    else if (match(cmdline, "HALF"))
        cols = SCRCOLS/2;
    else if ((cols = atoi(cmdline)) == 0)
        cols = SCRCOLS/2;
    wincol += cols;
    if (wincol < 0)
        wincol = 0;
    touched = TRUE;
}



