/*  select.c -- Define field constraints to select records for display

    The selection expression is converted from infix to postfix and is stored
    and evaluated on a set of stacks, according to the algorithm given in
    Horowitz and Sahni 1976, pp. 91-97 (who says that old dogs can't teach us
    new tricks ... or something like that :-).

    The function Postfix() converts the infix expression to an equivalent
    postfix expression. This is done by parsing 'tokens' from the expression.
    A token is a left or right parenthesis, a string delimited by single or
    double quotes, a string of alphanumeric characters, or a string of
    non-alphanumeric characters.

    The calling sequence for GetToken() is similar to strtok(), where the first
    call to the function is passed the address of the expression to parse and
    each subsequent call is passed a NULL pointer. Unlike strtok(), however,
    the original string is not changed.

    The algorithm for converting the infix expression to postfix is taken from
    that venerable old text, 'Fundamentals of Data Structures' Horowitz and
    Sahni, 1976. It is based on using an array to hold the final expression,
    and a stack to hold the intermediately parsed operators.

    Once the expression has been changed to postfix, evaluation is pretty
    simple.  The expression array is traversed, with operands pushed onto the
    stack and operators acting on stacked values.  The result of the evaluation
    is the value left on the top of the stack at the end.

    Written by Lee E. Brotzman, Hughes STX
    Written for Astronomical Data Center, NASA Goddard Space Flight Center

    This code is in the Public Domain, but no warranty, expressed or implied,
    is made by Hughes STX, NASA, or the ADC as to the accuracy and functioning
    of the software.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "ftb.h"
#include "screen.h"
#include "select.h"
#include "table.h"
#include "utility.h"

/*
    Global data, list of logical operator names, "AND", "OR", etc.  Used to
    print out selectio criteria.
*/
char  OPS[12][10];

/*
    Local data: stacks for evaluating expressions
*/
#define MAXCELLS 50
cell_t stack[MAXTABLES][MAXCELLS];      /* Postfix expression stacks */
int    top[MAXTABLES];                  /* Pointer to top-of-stack */
cell_t op_stack[MAXCELLS];              /* Stack for operators and operands */
int    op_top;                          /* Pointer to top-of-stack */

/*
    Local functions
*/
char  *GetToken(char *expression);
void   GetLogop(char *option, enum logop *operation, int *priority);
cell_t GetField(int tableno, int fieldno);
cell_t GetNull(int tableno, int fieldno);
char  *Value(int tableno, cell_t item);


void Select(char *cmdline)
/*
    Select a table and define the records that can be displayed.
*/
{   int  tableno;            /* Index of current table */

    SET.browse = FALSE;      /* Clear browse mode */

    GetTableNo(cmdline, &tableno);

    if (strlen(trim(cmdline)) == 0)
        PrintSel(tableno);
    else if (match(cmdline, "CLEAR"))
        top[tableno] = 0;
    else
        Postfix(tableno, cmdline);
}


void Postfix(int t, char *expression)
{
    char *token;
    enum logop op;
    int  priority;
    int  fieldno;
    int  i;
    char num[] = "0123456789+-.eEdD";

    table_t *T = table[t];    /* Shorthand */
    field_t *F;               /* Shorthand */

    op_top = top[t] = 0;
    token = GetToken(expression);
    for (;;) {
        /*
            If there are no more tokens, then clear the operand stack
        */
        if (token == NULL) {
            while (op_top) {
                op_top--;
                stack[t][top[t]++] = op_stack[op_top];
            }
            return;
        }
        /*
            If the token is a right parenthesis, pop the operand stack
            down to the corresponding left parenthesis
        */
        if (*token == ')') {
            while (op_top && op_stack[op_top-1].type != C_LPAREN) {
                op_top--;
                stack[t][top[t]++] = op_stack[op_top];
            }
            if (op_top)
                --op_top;
            goto next_token;
        }
        /*
            If the token is a left parenthesis, push it on the stack
        */
        if (*token == '(') {
            op_stack[op_top].type       = C_LPAREN;
            op_stack[op_top++].value.I  = 0;         /* Priority */
            goto next_token;
        }
        /*
            If the token is a logical operation, either push it to the
            operand stack, or the expression stack, depending on priority
        */
        GetLogop(token, &op, &priority);
        if (op != L_NONE) {
            while (op_top && op_stack[op_top-1].value.I >= priority) {
                op_top--;
                stack[t][top[t]++] = op_stack[op_top];
            }
            op_stack[op_top].type     = C_LOGOP;
            op_stack[op_top].operator = op;
            op_stack[op_top++].value.I= priority;
            goto next_token;
        }
        /*
            It's not a parenthesis or a logical operator, so it must be a field
            name or value
        */
        if (*token == '\'' || *token == '\"') { /* First check for a string */
            stack[t][top[t]].type      = C_CHAR;
            stack[t][top[t]++].value.A = strdup(token+1);
            goto next_token;
        }
        for (fieldno = i = 0; i <= T->fields; i++)  /* Check for a field */
            if ((F = T->field[i]) != NULL)
                if (stricmp(F->name, token) == 0) {
                    fieldno = i;
                    break;
                }
        if (fieldno) {                              /* It's a field */
            stack[t][top[t]].type      = C_FIELD;
            stack[t][top[t]++].value.I = fieldno;
        }
        else if (stricmp("NULL", token) == 0) {     /* It's a TNULL value */
            token = GetToken(NULL);                 /* Parse out '(field)' */
            if (*token != '(') {
                ErrMsg("Error: null() has unmatched parentheses");
                goto err_exit;
            }
            token = GetToken(NULL);
            for (fieldno = i = 0; i <= T->fields; i++)  /* Get field */
                if ((F = T->field[i]) != NULL)
                    if (stricmp(F->name, token) == 0) {
                        fieldno = i;
                        break;
                    }
            if (!fieldno) {
                ErrMsg("Error: invalid field name in null(), %s", token);
                goto err_exit;
            }
            else if (strlen(F->null) == 0) {
                ErrMsg("Error: field %s has no NULL value", token);
                goto err_exit;
            }
            token = GetToken(NULL);
            if (*token != ')') {
                ErrMsg("Error: null() has unmatched parentheses");
                goto err_exit;
            }
            stack[t][top[t]].type      = C_NULL;
            stack[t][top[t]++].value.I = fieldno;
        }
        else {                                      /* It's a numeric value */
            if (strspn(token, num) != strlen(token)) {
                ErrMsg("Error: unknown value, %s", token);
                goto err_exit;
            }
            stack[t][top[t]].type      = C_FLOAT;
            stack[t][top[t]++].value.F = atof(token);
        }

        /*
            Get the next token
        */
next_token:
        token = GetToken(NULL);
    }
err_exit:
    top[t] = 0;
}


char *GetToken(char *expression)
{
    static char *p;             /* Static pointer to expression */
    static char tok[81];        /* Buffer holding parsed token */
    char   *q = tok;            /* Dynamic output string pointer */
    char    quote;              /* Holder for quote character */

    if (expression)             /* If passed a non-NULL string pointer... */
        p = expression;         /* ...reset static pointer */

    if (*p == '\0')             /* Return NULL when at end-of-string */
        return NULL;

    while (isspace(*p))         /* Skip leading spaces */
        p++;

    /*
        Check for field name or number
    */
    if (isalnum(*p) || *p=='_' || *p=='+' || *p=='-' || *p=='.') {
        *q++ = *p++;
        while (isalnum(*p) || *p=='_' || *p=='+' || *p=='-' || *p=='.')
            *q++ = *p++;
    }
    /*
        Check for logical operator in mathmatical notation
    */
    else if (*p=='=' || *p=='>' || *p=='<' || *p=='!' || *p=='&' || *p=='|') {
        *q++ = *p++;
        while (*p=='=' || *p=='>' || *p=='<' || *p=='!' || *p=='&' || *p=='|')
            *q++ = *p++;
    }
    /*
        Check for quoted string
    */
    else if (*p=='\'' || *p=='\"') {
        quote = *p;
        *q++ = *p++;
        while (*p && *p != quote) {
            if (*p == '\\') p++;        /* Embedded quotes must be escaped */
            *q++ = *p++;
        }
        if (*p == quote)                /* Don't return closing quote */
            p++;
    }
    else                                /* Another symbol returned by itself */
        *q++ = *p++;

    *q = '\0';

    return tok;
}


void GetLogop(char *option, enum logop *operation, int *priority)
/*
    Determine which logical operation is defined by the given string.
    The priority is used by the expression routine to convert from infix to
    postfix notation on a stack.
*/
{
    if (!stricmp(option, "AND") ||
        !stricmp(option, "&&")  ||
        !stricmp(option, "&"))         { *operation = L_AND; *priority = 2; }

    else if (!stricmp(option, "OR") ||
             !stricmp(option, "||") ||
             !stricmp(option, "|"))    { *operation = L_OR; *priority = 1; }

    else if (!stricmp(option, "NOT") ||
             !stricmp(option, "!"))    { *operation = L_NOT; *priority = 4; }

    else if (!stricmp(option, "EQ") ||
             !stricmp(option, "==") ||
             !stricmp(option, "="))    { *operation = L_EQ; *priority = 3; }

    else if (!stricmp(option, "NE") ||
             !stricmp(option, "!=") ||
             !stricmp(option, "<>"))   { *operation = L_NE; *priority = 3; }

    else if (!stricmp(option, "LT") ||
             !stricmp(option, "<"))    { *operation = L_LT; *priority = 3; }

    else if (!stricmp(option, "GT") ||
             !stricmp(option, ">"))    { *operation = L_GT; *priority = 3; }

    else if (!stricmp(option, "GE") ||
             !stricmp(option, ">="))   { *operation = L_GE; *priority = 3; }

    else if (!stricmp(option, "LE") ||
             !stricmp(option, "<="))   { *operation = L_LE; *priority = 3; }

    else if (!stricmp(option, "SE"))   { *operation = L_SE; *priority = 3; }

    else if (!stricmp(option, "IN"))   { *operation = L_IN; *priority = 3; }

    else                               { *operation = L_NONE; *priority = 0; }

}


int Evaluate(int t)
{
    int i;

    if (top[t] == 0)                    /* Any selections for this table? */
        return TRUE;

    for (op_top = i = 0; i < top[t]; i++) {     /* Go through expression */
        switch (stack[t][i].type) {
            case C_CHAR: case C_FLOAT:          /* Push values */
                op_stack[op_top++] = stack[t][i];
                break;

            case C_FIELD:                       /* Convert field to value */
                op_stack[op_top++] = GetField(t, stack[t][i].value.I);
                break;

            case C_NULL:
                op_stack[op_top++] = GetNull(t, stack[t][i].value.I);
                break;

            case C_LOGOP:                       /* Perform operation */
                switch (stack[t][i].operator) {
                    case L_NOT:
                        op_stack[op_top-1].type = C_BOOL;
                        op_stack[op_top-1].value.I =
                            !(op_stack[op_top-1].value.I);
                        break;

                    case L_AND:
                        op_stack[op_top-2].type = C_BOOL;
                        op_stack[op_top-2].value.I =
                            op_stack[op_top-2].value.I &&
                            op_stack[op_top-1].value.I;
                        op_top--;
                        break;

                    case L_OR:
                        op_stack[op_top-2].type = C_BOOL;
                        op_stack[op_top-2].value.I =
                            op_stack[op_top-2].value.I ||
                            op_stack[op_top-1].value.I;
                        op_top--;
                        break;

                    case L_EQ :
                        op_stack[op_top-2].type = C_BOOL;
                        switch (op_stack[op_top-1].type) {
                            case C_FLOAT:
                                op_stack[op_top-2].value.I =
                                    op_stack[op_top-2].value.F ==
                                    op_stack[op_top-1].value.F;
                                break;
                            case C_CHAR:
                                op_stack[op_top-2].value.I =
                                    stricmp(op_stack[op_top-2].value.A,
                                            op_stack[op_top-1].value.A) == 0;
                                break;
                        }
                        op_top--;
                        break;
                    case L_NE :
                        op_stack[op_top-2].type = C_BOOL;
                        switch (op_stack[op_top-1].type) {
                            case C_FLOAT:
                                op_stack[op_top-2].value.I =
                                    op_stack[op_top-2].value.F !=
                                    op_stack[op_top-1].value.F;
                                break;
                            case C_CHAR:
                                op_stack[op_top-2].value.I =
                                    stricmp(op_stack[op_top-2].value.A,
                                            op_stack[op_top-1].value.A) != 0;
                                break;
                        }
                        op_top--;
                        break;
                    case L_LT :
                        op_stack[op_top-2].type = C_BOOL;
                        switch (op_stack[op_top-1].type) {
                            case C_FLOAT:
                                op_stack[op_top-2].value.I =
                                    op_stack[op_top-2].value.F <
                                    op_stack[op_top-1].value.F;
                                break;
                            case C_CHAR:
                                op_stack[op_top-2].value.I =
                                    stricmp(op_stack[op_top-2].value.A,
                                            op_stack[op_top-1].value.A) < 0;
                                break;
                        }
                        op_top--;
                        break;
                    case L_GT :
                        op_stack[op_top-2].type = C_BOOL;
                        switch (op_stack[op_top-1].type) {
                            case C_FLOAT:
                                op_stack[op_top-2].value.I =
                                    op_stack[op_top-2].value.F >
                                    op_stack[op_top-1].value.F;
                                break;
                            case C_CHAR:
                                op_stack[op_top-2].value.I =
                                    stricmp(op_stack[op_top-2].value.A,
                                            op_stack[op_top-1].value.A) > 0;
                                break;
                        }
                        op_top--;
                        break;
                    case L_LE :
                        op_stack[op_top-2].type = C_BOOL;
                        switch (op_stack[op_top-1].type) {
                            case C_FLOAT:
                                op_stack[op_top-2].value.I =
                                    op_stack[op_top-2].value.F <=
                                    op_stack[op_top-1].value.F;
                                break;
                            case C_CHAR:
                                op_stack[op_top-2].value.I =
                                    stricmp(op_stack[op_top-2].value.A,
                                            op_stack[op_top-1].value.A);
                                op_stack[op_top-2].value.I =
                                    (op_stack[op_top-2].value.I == 0) ||
                                    (op_stack[op_top-2].value.I < 0);
                                break;
                        }
                        op_top--;
                        break;
                    case L_GE :
                        op_stack[op_top-2].type = C_BOOL;
                        switch (op_stack[op_top-1].type) {
                            case C_FLOAT:
                                op_stack[op_top-2].value.I =
                                    op_stack[op_top-2].value.F >=
                                    op_stack[op_top-1].value.F;
                                break;
                            case C_CHAR:
                                op_stack[op_top-2].value.I =
                                    stricmp(op_stack[op_top-2].value.A,
                                            op_stack[op_top-1].value.A);
                                op_stack[op_top-2].value.I =
                                    (op_stack[op_top-2].value.I == 0) ||
                                    (op_stack[op_top-2].value.I > 0);
                                break;
                        }
                        op_top--;
                        break;
                    case L_SE :     /* Only valid for string comparisons */
                        op_stack[op_top-2].type = C_BOOL;
                        op_stack[op_top-2].value.I =
                            strcmp(op_stack[op_top-2].value.A,
                                   op_stack[op_top-1].value.A) == 0;
                        op_top--;
                        break;
                    case L_IN :     /* Only valid for string comparisons */
                        op_stack[op_top-2].type = C_BOOL;
                        op_stack[op_top-2].value.I =
                            strstr(strupr(op_stack[op_top-2].value.A),
                                   strupr(op_stack[op_top-1].value.A)) != NULL;
                        op_top--;
                        break;
                }

        }
    }
    return op_stack[op_top-1].value.I;
}


cell_t GetField(int tableno, int fieldno)
/*
    Get the value of a field that is required to perform the selection
    test and convert it to its binary equivalent.
*/
{
    int  source;        /* Index value of source string */
    int  count;         /* Number of bytes to move */
    int  digits;        /* Number of digits after '.' in F formats */
    char fldbuff[81];
    double readf();

    table_t *T = table[tableno];
    field_t *F = T->field[fieldno];
    cell_t  C;

    source = F->column - 1;
    count  = F->width;
    digits = F->digits;
    memcpy(fldbuff, &T->buffer[source], count);
    fldbuff[count] = '\0';

    switch (F->format[0]) {
        case 'I' :
            C.type    = C_FLOAT;
            C.value.F = atof(fldbuff);
            break;
        case 'F' : case 'D' : case 'E' :
            C.type    = C_FLOAT;
            C.value.F = readf(fldbuff, count, digits);
            break;
        case 'A' :
            C.type    = C_CHAR;
            C.value.A = strdup(fldbuff);
            break;
    }
    return C;
}


cell_t GetNull(int tableno, int fieldno)
/*
    Get the 'null' value of a field and compare to the value in the current
    record.  Store 'true' if they match, false otherwise.  All comparisons
    are exact string comparisons (to handle the case of non-numeric NULL
    values for numeric fields.
*/
{
    int  source;        /* Index value of source string */
    int  count;         /* Number of bytes to move */
    table_t *T = table[tableno];
    field_t *F = T->field[fieldno];
    cell_t  C;

    source = F->column - 1;     /* Point to the field in the data record */
    count  = F->width;
    C.type    = C_BOOL;
    C.value.I = !strncmp(F->null, &T->buffer[source], count);
    return C;
}


void PrintSel2(int t)
{
    int  i;
    table_t *T = table[t];

    Clear();
    if (top[t] == 0) {
        Say("No selection in effect.\n");
        return;
    }

    for (i = 0; i < top[t]; i++)
        if (stack[t][i].type == C_LOGOP)
            Say("Operator: %s\n", OPS[stack[t][i].operator]);
        else if (stack[t][i].type == C_FIELD)
            Say("Field   : %s\n", T->field[stack[t][i].value.I]->name);
        else if (stack[t][i].type == C_NULL)
            Say("Null()  : %s\n", T->field[stack[t][i].value.I]->name);
        else if (stack[t][i].type == C_BOOL)
            Say("Boolean : %s\n", stack[t][i].value.I ? "TRUE" : "FALSE");
        else if (stack[t][i].type == C_FLOAT)
            Say("FP value: %f\n", stack[t][i].value.F);
        else if (stack[t][i].type == C_CHAR)
            Say("String  : %s\n", stack[t][i].value.A);
        else
            Say("Unknown : ??\n");
}

void PrintSel(int t)
{
    int   wrap;
    char *op_stack[MAXCELLS];
    int   op_top;
    char *buff;
    char *p;
    int   i;

    if (top[t] == 0) {
        ErrMsg("WARNING: No selection in effect for table %d (%s)\n",
               t, table[t]->name);
        return;
    }

    Clear();
    wrap = SET.wrap;
    SET.wrap = TRUE;

    for (i = 0; i < MAXCELLS; i++)    /* Initialize temporary buffer stack */
        op_stack[i] = NULL;
    op_top = 0;
    buff = (char *)malloc(1000);

    /*
        This is actually pretty simple.  The postfix expression is stepped
        through from left to right.  All operands are stacked.  When an
        operator is encountered, the top-of-stack is the right hand operand,
        the cell under that is the left hand operand.  These are combined in
        infix notation, and pushed back onto the stack.  The one twist is the
        unary operator, NOT.  That only consumes one stack element.
    */
    for (i = 0; i < top[t]; i++) {
        if (stack[t][i].type != C_LOGOP) {   /* Stack operands */
            strcpy(buff, Value(t, stack[t][i]));
            if (op_stack[op_top])
                op_stack[op_top] =
                    (char *)realloc(op_stack[op_top], strlen(buff)+1);
            else
                op_stack[op_top] = (char *)malloc(strlen(buff)+1);
            strcpy(op_stack[op_top++], buff);
        }
        else if (stack[t][i].operator == L_NOT) {
            sprintf(buff, "%s(%s)",
                    OPS[stack[t][i].operator], op_stack[op_top-1]);
            op_stack[op_top-1] = (char *)realloc(op_stack[op_top-1],
                                                 strlen(buff)+1);
            strcpy(op_stack[op_top-1], buff);
        }
        else {
            sprintf(buff, "(%s %s %s)", op_stack[op_top-2],
                    OPS[stack[t][i].operator], op_stack[op_top-1]);
            op_stack[op_top-2] = (char *)realloc(op_stack[op_top-2],
                                                 strlen(buff)+1);
            strcpy(op_stack[op_top-2], buff);
            free(op_stack[op_top-1]);
            op_stack[op_top-1] = NULL;
            op_top--;
        }
    }
    /*
        Now there should be exactly one expression sitting on the stack.  The
        parentheses for this expression are redundant and are removed before
        printing.
    */
    if (op_top == 1) {
        p = strrchr(op_stack[0], ')');
        *p = '\0';
        Say("Selection setting for table %d (%s)\n\n", t, table[t]->name);
        Say("%s\n", op_stack[0]+1);
        free(op_stack[0]);
    }
    else {
        ErrMsg("ERROR: Invalid expression for SELECT condition\n");
        for (i = 0; i < MAXCELLS; i++)  /* Free temporary stack cells */
            if (op_stack[i])
                free(op_stack[i]);
    }
    free(buff);
    SET.wrap = wrap;
}

char *Value(int tableno, cell_t item)
{
static char buff[80];
    table_t *T = table[tableno];

    switch (item.type) {
        case C_FIELD:
            strcpy(buff, T->field[item.value.I]->name); break;
        case C_NULL:
            strcpy(buff, "Null(");
            strcat(buff, T->field[item.value.I]->name);
            strcat(buff, ")");  break;
        case C_BOOL:
            strcpy(buff, item.value.I ? "TRUE" : "FALSE"); break;
        case C_FLOAT:
            sprintf(buff, "%G", item.value.F); break;
        case C_CHAR:
            sprintf(buff, "\'%s\'", item.value.A); break;
        default:
            strcpy(buff, "(?? Unknown ??)");
    }
    return buff;
}


