/*  table.c -- Table structure initialization and search

    Table structure and function definitions.  The table structure holds all of
    the relavent descriptive information for the table being displayed. Some
    utilities are included for initializing a table structure, and allocating/
    initializing new dynamic elements.

    Written by Lee E. Brotzman, Hughes STX
    Written for Astronomical Data Center, NASA Goddard Space Flight Center

    This code is in the Public Domain, but no warranty, expressed or implied,
    is made by Hughes STX, NASA, or the ADC as to the accuracy and functioning
    of the software.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include "ftb.h"
#include "screen.h"
#include "table.h"
#include "utility.h"


/*
    Allocate the table data structures
*/
file_t  *file[MAXFILES];    /* File descriptor data structure */
table_t *table[MAXTABLES];  /* Table descriptor data structure */


comment_t *FreeList(comment_t *P)
/*
    Free all nodes on the circularly-linked list of comments and history.
    The list pointer points to the last element in the list.
*/
{   comment_t  *Q, *R;

    Q = P->next;        /* Start with the first element in the list */
    while (Q != P) {    /* Go through and get rid of all nodes */
        R = Q;
        Q = Q->next;
        free(R);
    }
    free(Q);            /* Free the last node */
    return NULL;        /* Ensure that P is NULL */
}


void FreeTable(int tableno)
/*
    Free and/or initialize all pieces of the table structure.  READ THIS
    FUNCTION CAREFULLY, for from this code certain assumptions about
    what the data structures look like are made.
*/
{   int     i;      /* Loop control variables */
    table_t *T = table[tableno];    /* Shorthand */

    if (T != NULL) {
        if (T->buffer != NULL)
            free(T->buffer);
        if (T->comment != NULL)
            T->comment = FreeList(T->comment);
        if (T->history != NULL)
            T->history = FreeList(T->history);
        for (i = 0; i < MAXFIELDS; i++)
            if (T->field[i] != NULL) {
                if (T->field[i]->comment != NULL)
                    T->field[i]->comment = FreeList(T->field[i]->comment);
                free(T->field[i]);
            }
        free(T);
    }

    for (i = tableno+1; i <= TABLES; i++)
        table[i-1] = table[i];
    table[i-1] = NULL;
    if (SET.table == TABLES)
        SET.table--;
    TABLES--;                           /* Decrement global table count */
}


void FreeFile(int filenum)
/*
    Free and/or initialize all pieces of the file structure.  READ THIS
    FUNCTION CAREFULLY, for from this code certain assumptions about
    what the data structures look like are made.
*/
{   file_t *F = file[filenum];  /* Shorthand */

    if (F != NULL)
        if (--(F->links) <= 0) {
            if (F->filename != NULL)
                free(F->filename);
            if (F->comment != NULL)
                F->comment = FreeList(F->comment);
            fclose(F->fileptr);
            free(F);
            file[filenum] = NULL;
            FILES--;                    /* Decrement global file count */
        }
}


table_t *NewTable(void)
/*
    Create and initialize a table entry pointed to by T.    READ THIS
    FUNCTION CAREFULLY, for from this code certain assumptions about
    what the data structures look like are made.
*/
{   table_t *T;
    int     i;

    if ((T = (table_t *) malloc(sizeof(table_t))) != NULL) {
        T->filenum      = 0;
        T->hdrbase      = 0;
        T->tabbase      = 0;
        T->buffer       = NULL;
        T->cols         = 0;
        T->rows         = 0;
        T->fields       = 0;
        T->name[0]      = '\0';
        T->namedesc[0]  = '\0';
        T->author[0]    = '\0';
        T->reference[0] = '\0';
        T->date[0]      = '\0';
        T->comment      = NULL;
        T->history      = NULL;
        for (i = 0; i < MAXFIELDS; i++)
            T->field[i] = NULL;
    }
    return T;
}


field_t *NewField(void)
/*
    Create and initialize a field entry pointed to by F.    READ THIS
    FUNCTION CAREFULLY, for from this code certain assumptions about
    what the data structures look like are made.
*/
{   field_t *F;

    if ((F = (field_t *) malloc(sizeof(field_t))) != NULL) {
        F->name[0]      = '\0';
        F->namedesc[0]  = '\0';
        F->column       = 0;
        F->width        = 0;
        F->digits       = 0;
        F->format[0]    = '\0';
        F->cfmt[0]      = '\0';
        F->units[0]     = '\0';
        F->unitdesc[0]  = '\0';
        F->null[0]      = '\0';
        F->scale        = 1.0;
        F->zeropt       = 0.0;
        F->comment      = NULL;
    }
    return F;
}


file_t *NewFile(void)
/*
    Create and initialize a file entry pointed to by F.  READ THIS
    FUNCTION CAREFULLY, for from this code certain assumptions about
    what the data structures look like are made.
*/
{   file_t *F;

    if ((F = (file_t *) malloc(sizeof(file_t))) != NULL) {
        FILES++;
        F->filename     = NULL;
        F->fileptr      = NULL;
        F->filetype     = FTB_UNK;
        F->links        = 0;
        F->comment      = NULL;
    }
    return F;
}


void FieldID(int tableno, char *fldname, int *fieldno)
/*
    Convert a field identifier string into the corresponding field number.
    The field identifier can be either a numeric string or the field name.
    If the field identifier is invalid, return with fieldno = 0. (Note: the
    field number is always returned counted from 1, because they are stored
    that way.)
*/
{   int     i;  /* Loop control variable */
    table_t *T = table[tableno];    /* Shorthand */
    field_t *F;                     /* Shorthand */

    /*
        Look for the name first, to guard against screwing up if a field has a
        number as the first character(s), e.g. 1ST_FIELD.
    */
    for (*fieldno = i = 0; i <= T->fields; i++)
        if ((F = T->field[i]) != NULL)
            if (strlen(F->name) == strlen(fldname) &&
                strnicmp(F->name, fldname, strlen(fldname)) == 0) {
                    *fieldno = i;
                    break;
                }

    if (*fieldno == 0) {        /* Test for numeric string */
        *fieldno = atoi(fldname);
        if (*fieldno < 0 || *fieldno > T->fields)
            *fieldno = 0;
    }
}


void TableID(char *tabname, int *tableno)
/*
    Convert a table identifier string into the corresponding table number.
    The table identifier can be either a numeric string or the table name.
    If the table identifier is invalid, return with tableno = 0 (Note: the
    table number is always counted from 1).
*/
{   int i;       /* Loop control variable */

    /*
        Look for a name first to guard against screwing up if a table has a
        number as the first character(s), e.g. 3RD_TABLE.
    */
    for (*tableno = i = 0; i < MAXTABLES; i++)
        if (table[i] != NULL)   /* Is there a table here? */
            if (strnicmp(table[i]->name, tabname, strlen(tabname)) == 0) {
                 *tableno = ++i;
                 break;
            }

    if (*tableno == 0) {                /* Test for numeric string */
        *tableno = atoi(tabname);
        if (*tableno <= 0 || *tableno > MAXTABLES)
            *tableno = 0;
        if (table[*tableno] == NULL)    /* Is there a table here? */
            *tableno = 0;
    }
}


void GetTableNo(char *cmdline, int *tableno)
/*
    Parse a table identifier from the command line, if the "TABLE" option
    is given.  If an illegal table identifier is given, then generate a
    message and return the current table number.
*/
{   char *option;       /* Temporary string for scanning command line */
    char *backup;       /* Temporary string for backup of command line */
    option = strdup(cmdline);
    backup = strdup(cmdline);

    scan(cmdline, ' ', option);
    if (match(option, "TABLE")) {   /* If "Table" found... */
        scan(cmdline, ' ', option);
        TableID(option, tableno);   /* Convert table name/number to an index */
        if (*tableno != 0)
            SET.table = *tableno;
        else {
            *tableno = SET.table;
            ErrMsg("Invalid table identifier entered: %s", option);
            if (table[*tableno] != NULL)
                ErrMsg("Default table (%d. %s) assumed.",
                        *tableno, table[*tableno]->name);
        }
    }
    else {                          /* If "Table" not found... */
        *tableno = SET.table;       /* Use default table index */
        strcpy(cmdline, backup);    /* Restore the original command line */
    }

    free(option);   /* Release the temporary strings */
    free(backup);

    if (*tableno == 0 || table[*tableno] == NULL) {
        ErrMsg("Table number %d is invalid, open or select a new table",
                *tableno);
        longjmp(CmdJmp, 0);
    }
}
