/*  table.h -- Table structure header file

    The table structure holds all the relevant descriptive information for the
    table being displayed. Some utilities are included for initializing a table
    structure, and allocating/initializing new dynamic elements.

    Written by Lee E. Brotzman, Hughes STX
    Written for Astronomical Data Center, NASA Goddard Space Flight Center

    This code is in the Public Domain, but no warranty, expressed or implied,
    is made by Hughes STX, NASA, or the ADC as to the accuracy and functioning
    of the software.
*/
#ifndef __TABLE_H
#define __TABLE_H

#include <stdio.h>

/*
    Define lengths of various character strings in the table structure
*/
#define SHORTSTR 21
#define MEDSTR 51
#define LONGSTR 73

/*
    Define enumerations for the various file types accessed by FTB: unknown,
    ordinary text file, Standard FITS ASCII table, text file FITS table header,
    text file catalog data, FITS binary table (not yet implemented).
*/
typedef enum {
    FTB_UNK, FTB_TXT, FTB_FITS, FTB_HDR, FTB_DAT, FTB_BIN
}   ftb_t;


/*
    Comments and history are kept in a circularly-linked list; the list
    pointer points to the last record in the list.
*/
typedef struct com_node {
    struct  com_node *next;
    char    data[LONGSTR];
}   comment_t;


/*
    Open files are described by the fully qualified name, file pointer, and
    the number of links to that file.
*/
typedef struct {
    char      *filename;    /* Name of file */
    FILE      *fileptr;     /* System file pointer */
    ftb_t      filetype;    /* File type (see above) */
    int        links;       /* Number of links to file */
    comment_t *comment;     /* List of COMMENT records in main header */
}   file_t;


/*
    The field structure holds information about individual fields in the table.
*/
typedef struct {
    char    name[LONGSTR];      /* Name of field */
    char    namedesc[MEDSTR];   /* English description of field name */
    int     column;             /* Starting column */
    int     width;              /* Width of field */
    int     digits;             /* Digits after '.' (F,E,D formats) */
    char    format[SHORTSTR];   /* FORTRAN format of the field */
    char    cfmt[SHORTSTR];     /* C format of the field */
    char    units[LONGSTR];     /* Units for data */
    char    unitdesc[MEDSTR];   /* English description of units */
    char    null[LONGSTR];      /* Null value */
    double  scale;              /* Scale factor */
    double  zeropt;             /* Zero point offset value */
comment_t   *comment;           /* Comment */
}   field_t;


/*
    The table structure holds overall descriptive information for an individual
    table, including an array of pointers to field descriptors.
*/
typedef struct {                /* Define structure for individual table */
    int     filenum;            /* Index to file array */
    long    hdrbase;            /* File offset to first header record */
    long    tabbase;            /* File offset to first table record */
    char    *buffer;            /* Dynamically allocated input buffer */
    int     cols;               /* Number of characters/record */
    long    rows;               /* Number of records/table */
    int     fields;             /* Number of fields/record */
    char    name[LONGSTR];      /* Name of table */
    char    namedesc[MEDSTR];   /* English expansion of the name */
    char    author[LONGSTR];    /* Author(s) */
    char    reference[LONGSTR]; /* Reference */
    char    date[SHORTSTR];     /* Date table was generated */
    comment_t *comment;         /* Pointer to comments circular list */
    comment_t *history;         /* Pointer to history circular list */
    field_t   *field[MAXFIELDS];/* Pointers to field information */
}   table_t;


/*
    External data definitions and function prototypes
*/
extern file_t   *file[MAXFILES];    /* Global file data structure */
extern table_t  *table[MAXTABLES];  /* Global table data structure */

comment_t   *FreeList(comment_t *comment); /* Free linked list of comments */
void        FreeTable(int tableno);     /* Free dynamic elements in table[] */
void        FreeFile(int filenum);      /* Free dynamic elements in file[] */
table_t     *NewTable(void);            /* Initialize new table entry */
field_t     *NewField(void);            /* Initialize new field entry */
file_t      *NewFile(void);             /* Initialize new file entry */

        /* Convert table and field identifiers */
void    TableID(char *tabname, int *tableno);
void    FieldID(int tableno, char *fldname, int *fieldno);
void    GetTableNo(char *cmdline, int *tableno);

#endif
