/*  utility.c -- String utilities

    This file contains a few utility functions for string manipulation.

    Written by Lee E. Brotzman, Hughes STX
    Written for Astronomical Data Center, NASA Goddard Space Flight Center

    This code is in the Public Domain, but no warranty, expressed or implied,
    is made by Hughes STX, NASA, or the ADC as to the accuracy and functioning
    of the software.
*/
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "ftb.h"
#include "utility.h"


#ifndef MSDOS
char *strupr(char *s)
/*
    Return pointer to upper case string
*/
{   char *t;

    for (t = s; *t; t++)
        *t = toupper(*t);
    return s;
}
#endif

#ifdef ultrix
char *strdup(char *s)
/*
    Return a pointer to a dynamically-allocated duplicate of the input string.
*/
{   char *t;

    if ((t = (char *)malloc(strlen(s)+1)) == NULL) {
        Clear();
        FreeScr();
        printf("strdup: error allocating string\n");
        exit(1);
    }
    strcpy(t, s);
    return t;
}
#endif

char *skip(char *s, char delim)
/*
    Skip all leading delimiter characters in s and return the modified string.
*/
{   char *t;

    /* Skip all leading delimiters */
    for (t = s; *t == delim; t++)
         continue;

    /* Set s to be the string without leading delimiters */
    strcpy(s, t);
    return s;
}


char *trim(char *s)
/*
    Trim leading and trailing blanks from a string.
*/
{   int   i;
    char *t;

    /* Trim leading blanks */
    for (t = s; isspace(*t); t++)
        continue;
    strcpy(s, t);

    /* Trim trailing blanks */
    for (i = strlen(s)-1; (isspace(s[i])) && (i >= 0); i--)
        continue;

    s[i+1] = '\0';
    return s;
}


char *scan(char *s, char delim, char *object)
/*
    Parse the next object in s.  An object is a string which is delimited
    by the given delimiter character and/or the boundary of s (if the object is
    at the very beginning or the very end of s).  s is modified to contain the
    string following the object.  NOTE: if the delimiter is an ASCII blank,
    what I really want to parse on is white space.  Therefore, there are
    two separate but equal algorithms in this function, one for whitespace,
    one for any other kind of delimiter.
*/
{   char *t;

    if (delim == ' ') {
        /* Skip any leading white space */
        for(t = s; isspace(*t); t++)
            continue;
        /* Parse through, and capture, any non-white space */
        for( ; *t && !isspace(*t); t++, object++)
            *object = *t;
        *object = '\0';
        /* Skip any following white space */
        for( ; isspace(*t); t++)
            continue;
    }
    else {
        /* Skip any leading delimiters */
        for(t = s; *t == delim; t++)
            continue;
        /* Parse through, and capture, any non-delimiters */
        for( ; *t && (*t != delim); t++, object++)
            *object = *t;
        *object = '\0';
        /* Skip any following delimiters */
        for( ; *t == delim; t++)
            continue;
    }
    /* Set s to be the remainder of the string */
    strcpy(s, t);
    return s;
}


int match(char *target, char *test)
/*
    The 'test' string is matched against the 'target' string.  The test string
    must match (case insensitively) the target up to the end of the target.
*/
{   /*
        Test for exception, i.e. null target string NEVER matches.
    */
    if (*target == '\0' && *test != '\0')
        return FALSE;

    /*
        Match until end-of-string is found
    */
    for ( ; *target; target++, test++)
        if (*test == '\0' || toupper(*target) != toupper(*test))
            return FALSE;

    return TRUE;
}
